from woke.utils import StrEnum


class RequestMethodEnum(StrEnum):
    # General
    INITIALIZE = "initialize"
    INITIALIZED = "initialized"  # Notification
    SHUTDOWN = "shutdown"
    EXIT = "exit"  # Notification

    # Window
    WINDOW_SHOW_MESSAGE = "window/showMessage"  # Notification
    WINDOW_SHOW_MESSAGE_REQUEST = "window/showMessageRequest"
    WINDOW_SHOW_DOCUMENT = "window/showDocument"
    WINDOW_LOG_MESSAGE = "window/logMessage"  # Notification
    WINDOW_WORK_DONE_PROGRESS_CREATE = "window/workDoneProgress/create"
    WINDOW_WORK_DONE_PROGRESS_CANCEL = "window/workDoneProgress/cancel"  # Notification

    # Telemetry
    TELEMETRY_EVENT = "telemetry/event"  # Notification

    # Client
    CLIENT_REGISTER_CAPABILITY = "client/registerCapability"
    CLIENT_UNREGISTER_CAPABILITY = "client/unregisterCapability"

    # Workspace
    WORKSPACE_SYMBOL = "workspace/symbol"
    WORKSPACE_SYMBOL_RESOLVE = "workspaceSymbol/resolve"
    WORKSPACE_CONFIGURATION = "workspace/configuration"
    WORKSPACE_DID_CHANGE_CONFIGURATION = (
        "workspace/didChangeConfiguration"  # Notification
    )
    WORKSPACE_WORKSPACE_FOLDERS = "workspace/workspaceFolders"
    WORKSPACE_DID_CHANGE_WORKSPACE_FOLDERS = (
        "workspace/didChangeWorkspaceFolders"  # Notification
    )
    WORKSPACE_EXECUTE_COMMAND = "workspace/executeCommand"
    WORKSPACE_APPLY_EDIT = "workspace/applyEdit"
    WORKSPACE_CODE_LENS_REFRESH = "workspace/codeLens/refresh"
    WORKSPACE_DIAGNOSTIC = "workspace/diagnostic"
    WORKSPACE_DIAGNOSTIC_REFRESH = "workspace/diagnostic/refresh"

    # File Operations
    WORKSPACE_WILL_CREATE_FILES = "workspace/willCreateFiles"
    WORKSPACE_DID_CREATE_FILES = "workspace/didCreateFiles"  # Notification
    WORKSPACE_WILL_RENAME_FILES = "workspace/willRenameFiles"
    WORKSPACE_DID_RENAME_FILES = "workspace/didRenameFiles"  # Notification
    WORKSPACE_WILL_DELETE_FILES = "workspace/willDeleteFiles"
    WORKSPACE_DID_DELETE_FILES = "workspace/didDeleteFiles"  # Notification
    WORKSPACE_DID_CHANGE_WATCHED_FILES = (
        "workspace/didChangeWatchedFiles"  # Notification
    )

    # Text Synchronization
    TEXT_DOCUMENT_DID_OPEN = "textDocument/didOpen"  # Notification
    TEXT_DOCUMENT_DID_CHANGE = "textDocument/didChange"  # Notification
    TEXT_DOCUMENT_WILL_SAVE = "textDocument/willSave"  # Notification
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL = "textDocument/willSaveWaitUntil"
    TEXT_DOCUMENT_DID_SAVE = "textDocument/didSave"  # Notification
    TEXT_DOCUMENT_DID_CLOSE = "textDocument/didClose"  # Notification

    # Notebook Document
    NOTEBOOK_DOCUMENT_DID_OPEN = "notebookDocument/didOpen"  # Notification
    NOTEBOOK_DOCUMENT_DID_CHANGE = "notebookDocument/didChange"  # Notification
    NOTEBOOK_DOCUMENT_DID_SAVE = "notebookDocument/didSave"  # Notification
    NOTEBOOK_DOCUMENT_DID_CLOSE = "notebookDocument/didClose"  # Notification

    # Language Features
    DECLARATION = "textDocument/declaration"
    DEFINITION = "textDocument/definition"
    TYPE_DEFINITION = "textDocument/typeDefinition"
    IMPLEMENTATION = "textDocument/implementation"
    REFERENCES = "textDocument/references"
    PREPARE_CALL_HIERARCHY = "textDocument/prepareCallHierarchy"
    CALL_HIERARCHY_INCOMING_CALLS = "callHierarchy/incomingCalls"
    CALL_HIERARCHY_OUTGOING_CALLS = "callHierarchy/outgoingCalls"
    PREPARE_TYPE_HIERARCHY = "textDocument/prepareTypeHierarchy"
    TYPE_HIERARCHY_SUPERTYPES = "typeHierarchy/supertypes"
    TYPE_HIERARCHY_SUBTYPES = "typeHierarchy/subtypes"
    DOCUMENT_HIGHLIGHT = "textDocument/documentHighlight"
    DOCUMENT_LINK = "textDocument/documentLink"
    DOCUMENT_LINK_RESOLVE = "documentLink/resolve"
    HOVER = "textDocument/hover"
    CODE_LENS = "textDocument/codeLens"
    CODE_LENS_RESOLVE = "codeLens/resolve"
    FOLDING_RANGE = "textDocument/foldingRange"
    SELECTION_RANGE = "textDocument/selectionRange"
    DOCUMENT_SYMBOL = "textDocument/documentSymbol"
    SEMANTIC_TOKENS = "textDocument/semanticTokens"  # TODO: only used for registration
    SEMANTIC_TOKENS_FULL = "textDocument/semanticTokens/full"
    SEMANTIC_TOKENS_FULL_DELTA = "textDocument/semanticTokens/full/delta"
    SEMANTIC_TOKENS_RANGE = "textDocument/semanticTokens/range"
    SEMANTIC_TOKENS_REFRESH = "workspace/semanticTokens/refresh"
    INLAY_HINT = "textDocument/inlayHint"
    INLAY_HINT_RESOLVE = "inlayHint/resolve"
    INLAY_HINT_REFRESH = "workspace/inlayHint/refresh"
    INLINE_VALUE = "textDocument/inlineValue"
    INLINE_VALUE_REFRESH = "textDocument/inlineValue/refresh"
    MONIKER = "textDocument/moniker"
    COMPLETION = "textDocument/completion"
    COMPLETION_ITEM_RESOLVE = "completionItem/resolve"
    PUBLISH_DIAGNOSTICS = "textDocument/publishDiagnostics"  # Notification
    DIAGNOSTIC = "textDocument/diagnostic"
    SIGNATURE_HELP = "textDocument/signatureHelp"
    CODE_ACTION = "textDocument/codeAction"
    CODE_ACTION_RESOLVE = "codeAction/resolve"
    DOCUMENT_COLOR = "textDocument/documentColor"
    COLOR_PRESENTATION = "textDocument/colorPresentation"
    FORMATTING = "textDocument/formatting"
    RANGE_FORMATTING = "textDocument/rangeFormatting"
    ON_TYPE_FORMATTING = "textDocument/onTypeFormatting"
    RENAME = "textDocument/rename"
    PREPARE_RENAME = "textDocument/prepareRename"
    LINKED_EDITING_RANGE = "textDocument/linkedEditingRange"

    # Other
    CANCEL_REQUEST = "$/cancelRequest"
    PROGRESS = "$/progress"
    LOG_TRACE = "$/logTrace"  # Notification
    SET_TRACE = "$/setTrace"  # Notification



from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple

if TYPE_CHECKING:
    from woke.lsp.context import LspContext
    from woke.lsp.server import LspServer

from woke.compiler.solc_frontend import SolcOutputError, SolcOutputErrorSeverityEnum
from woke.lsp.common_structures import (
    Diagnostic,
    DiagnosticSeverity,
    DocumentUri,
    Position,
    Range,
)
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.methods import RequestMethodEnum
from woke.lsp.utils.uri import path_to_uri

logger = logging.getLogger(__name__)


class PublishDiagnosticsParams(LspModel):
    uri: DocumentUri
    """
    The URI for which diagnostic information is reported.
    """
    version: Optional[int] = None
    """
    Optiona the version number of the document the diagnostics are published
    for.
    
    @since 3.15.0
    """
    diagnostics: List[Diagnostic]
    """
    An array of diagnostic information items.
    """


async def diagnostics_loop(server: LspServer, context: LspContext):
    queue: asyncio.Queue = context.diagnostics_queue
    while True:
        file, diagnostics = await queue.get()

        params = PublishDiagnosticsParams(
            uri=DocumentUri(path_to_uri(file)),
            diagnostics=list(diagnostics),
        )
        await server.send_notification(RequestMethodEnum.PUBLISH_DIAGNOSTICS, params)



from __future__ import annotations

from typing import Iterator, Union

from woke.ir.ast import (
    SolcYulForLoop,
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulLiteral,
)

from ..utils import IrInitTuple
from .abc import YulAbc, YulStatementAbc
from .block import YulBlock
from .function_call import YulFunctionCall
from .identifier import YulIdentifier
from .literal import YulLiteral


class YulForLoop(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _body: YulBlock
    _condition: Union[YulFunctionCall, YulIdentifier, YulLiteral]
    _post: YulBlock
    _pre: YulBlock

    def __init__(self, init: IrInitTuple, for_loop: SolcYulForLoop, parent: YulAbc):
        super().__init__(init, for_loop, parent)
        self._body = YulBlock(init, for_loop.body, self)
        if isinstance(for_loop.condition, SolcYulFunctionCall):
            self._condition = YulFunctionCall(init, for_loop.condition, self)
        elif isinstance(for_loop.condition, SolcYulIdentifier):
            self._condition = YulIdentifier(init, for_loop.condition, self)
        elif isinstance(for_loop.condition, SolcYulLiteral):
            self._condition = YulLiteral(init, for_loop.condition, self)
        else:
            assert False, f"Unexpected type: {type(for_loop.condition)}"
        self._post = YulBlock(init, for_loop.post, self)
        self._pre = YulBlock(init, for_loop.pre, self)

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._pre
        yield from self._condition
        yield from self._body
        yield from self._post

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def body(self) -> YulBlock:
        return self._body

    @property
    def condition(self) -> Union[YulFunctionCall, YulIdentifier, YulLiteral]:
        return self._condition

    @property
    def post(self) -> YulBlock:
        return self._post

    @property
    def pre(self) -> YulBlock:
        return self._pre



from __future__ import annotations

import logging
from typing import List, Optional, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)

logger = logging.getLogger(__name__)
_recursion_guard = set()


def _process_assigned_vars(address_balance: ir.ExpressionAbc) -> List[Detection]:
    ret = []
    if address_balance in _recursion_guard:
        return ret
    _recursion_guard.add(address_balance)

    assigned_var_statement = address_balance
    parents: List[ir.IrAbc] = [address_balance]
    assigned_vars: Optional[Set[ir.AssignedVariablePath]] = None
    statement = None
    while assigned_var_statement is not None:
        if isinstance(assigned_var_statement, ir.Assignment):
            statement = assigned_var_statement
            assigned_variables = assigned_var_statement.assigned_variables
            if len(assigned_variables) > 1:
                parent = parents[-1]
                assert isinstance(parent, ir.TupleExpression)
                index = parent.components.index(parents[-2])
                assigned_vars = assigned_variables[index]
            else:
                assigned_vars = assigned_variables[0]
            break
        elif isinstance(assigned_var_statement, ir.VariableDeclarationStatement):
            statement = assigned_var_statement
            if len(assigned_var_statement.declarations) > 1:
                parent = parents[-1]
                assert isinstance(parent, ir.TupleExpression)
                index = parent.components.index(parents[-2])
                decl = assigned_var_statement.declarations[index]
                assert decl is not None
                assigned_vars = {(decl,)}
            else:
                decl = assigned_var_statement.declarations[0]
                assert decl is not None
                assigned_vars = {(decl,)}
            break
        elif isinstance(assigned_var_statement, ir.FunctionCall):
            break
        elif isinstance(assigned_var_statement, ir.FunctionCallOptions):
            break
        elif isinstance(assigned_var_statement, ir.BinaryOperation):
            if assigned_var_statement.operator in {
                ir.enums.BinaryOpOperator.EQ,
                ir.enums.BinaryOpOperator.NEQ,
            }:
                ret.append(
                    Detection(
                        assigned_var_statement,
                        f"Strict comparison on expression originating from address.balance",
                    )
                )

        parents.append(assigned_var_statement)
        assigned_var_statement = assigned_var_statement.parent

    if assigned_vars is None:
        return ret

    assert statement is not None
    for var in assigned_vars:
        for segment in var:
            if isinstance(segment, ir.VariableDeclaration):
                if segment.is_state_variable:
                    ret.append(Detection(statement, "State variable assignment here"))

    if len(assigned_vars) > 1 or len(next(iter(assigned_vars))) > 1:
        # currently not supported
        logger.debug(f"Complex assignment not supported: {statement.source}")
        return ret

    assigned_var = next(iter(assigned_vars))[0]
    assert assigned_var != "IndexAccess" and not isinstance(assigned_var, ir.SourceUnit)
    assigned_var_statement = None
    function_def = None
    node = assigned_var
    while node is not None:
        if isinstance(node, ir.StatementAbc) and assigned_var_statement is None:
            assigned_var_statement = node
        if isinstance(node, ir.FunctionDefinition):
            function_def = node
            break
        node = node.parent
    if assigned_var_statement is None:
        logger.debug(f"Could not find statement for {assigned_var.source}")
        return ret
    if function_def is None:
        logger.debug(f"Could not find function definition for {assigned_var.source}")
        return ret

    cfg = function_def.cfg
    assert cfg is not None

    for ref in assigned_var.references:
        if isinstance(ref, (ir.Identifier, ir.MemberAccess)):
            ref_statement = ref
            while ref_statement is not None:
                if isinstance(ref_statement, ir.StatementAbc):
                    break
                ref_statement = ref_statement.parent
            if ref_statement is None or ref_statement == assigned_var_statement:
                logger.debug(f"Could not find statement for {ref.source}")
                continue

            if cfg.is_reachable(assigned_var_statement, ref_statement):
                ret.extend(_process_assigned_vars(ref))
    return ret


class UnsafeAddressBalanceUseDetector(Detector):

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        print(ir.enums.GlobalSymbolsEnum.ADDRESS_BALANCE)
        return list(self._detections)

    def visit_member_access(self, node):
        if node.referenced_declaration == ir.enums.GlobalSymbolsEnum.ADDRESS_BALANCE:
            detections = _process_assigned_vars(node)
            if len(detections) > 0:
                self._detections.add(
                    DetectorResult(
                        Detection(
                            node, "Unsafe use of address.balance", tuple(detections)
                        ),
                        DetectionImpact.MEDIUM,
                        DetectionConfidence.MEDIUM,
                    ),
                )

    @detector.command("unsafe-address-balance-use")
    def unsafe_address_balance_use(self):
        """
        Address.balance is either written to a state variable or used in a strict comparison (== or !=).
        """
        pass



import logging
import re
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, List, Optional, Union

from woke.ir import (
    BinaryOperation,
    DeclarationAbc,
    ExternalReference,
    Identifier,
    IdentifierPath,
    IdentifierPathPart,
    IrAbc,
    MemberAccess,
    UnaryOperation,
    UserDefinedTypeName,
    YulIdentifier,
)
from woke.lsp.common_structures import (
    DocumentUri,
    OptionalVersionedTextDocumentIdentifier,
    Range,
    TextDocumentEdit,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    TextEdit,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
    WorkspaceEdit,
)
from woke.lsp.context import LspContext
from woke.lsp.exceptions import LspError
from woke.lsp.protocol_structures import ErrorCodes
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path

logger = logging.getLogger(__name__)


class RenameOptions(WorkDoneProgressOptions):
    prepare_provider: Optional[bool]
    """
    Renames should be checked and tested before being executed.
    """


class RenameRegistrationOptions(TextDocumentRegistrationOptions, RenameOptions):
    pass


class RenameParams(TextDocumentPositionParams, WorkDoneProgressParams):
    new_name: str
    """
    The new name of the symbol. If the given name is not valid the
    request must return a [ResponseError](#ResponseError) with an
    appropriate message set.
    """


class PrepareRenameParams(TextDocumentPositionParams, WorkDoneProgressParams):
    pass


def _generate_reference_location(
    reference: Union[
        Identifier,
        IdentifierPathPart,
        MemberAccess,
        ExternalReference,
        DeclarationAbc,
    ],
    context: LspContext,
) -> Range:
    path = reference.file
    if isinstance(reference, MemberAccess):
        location = reference.member_byte_location
    elif isinstance(reference, ExternalReference):
        location = reference.identifier_byte_location
    elif isinstance(reference, DeclarationAbc):
        location = reference.name_location
    else:
        location = reference.byte_location
    return context.compiler.get_range_from_byte_offsets(path, location)


def _generate_workspace_edit(
    declaration: DeclarationAbc, new_name: str, context: LspContext
) -> WorkspaceEdit:
    changes_by_file: DefaultDict[Path, List[TextEdit]] = defaultdict(list)

    for reference in declaration.get_all_references(True):
        if not isinstance(reference, (UnaryOperation, BinaryOperation)):
            changes_by_file[reference.file].append(
                TextEdit(
                    range=_generate_reference_location(reference, context),
                    new_text=new_name,
                )
            )

    changes: DefaultDict[DocumentUri, List[TextEdit]] = defaultdict(list)
    document_changes = []
    for file, edits in changes_by_file.items():
        changes[DocumentUri(path_to_uri(file))].extend(edits)
        document_changes.append(
            TextDocumentEdit(
                text_document=OptionalVersionedTextDocumentIdentifier(
                    version=context.compiler.get_compiled_file(file).version,
                    uri=DocumentUri(path_to_uri(file)),
                ),
                edits=edits,
            )
        )

    return WorkspaceEdit(
        document_changes=document_changes,
    )


IDENTIFIER_RE = re.compile(r"^[a-zA-Z$_][a-zA-Z0-9$_]*$")


async def rename(
    context: LspContext,
    params: RenameParams,
) -> Union[WorkspaceEdit, None]:
    """
    Renames the symbol denoted by the given text document position.
    """
    logger.debug(
        f"Requested rename for file {params.text_document.uri} at {params.position}"
    )

    match = IDENTIFIER_RE.match(params.new_name)
    if not match:
        raise LspError(ErrorCodes.InvalidRequest, "Invalid identifier name")

    logger.debug(f"Waiting for compiler")

    await context.compiler.output_ready.wait()

    logger.debug(f"Waiting done")

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, DeclarationAbc):
        name_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.name_location
        )
        if not position_within_range(params.position, name_location_range):
            raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")
    elif isinstance(node, Identifier):
        node = node.referenced_declaration
    elif isinstance(node, MemberAccess):
        member_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.member_byte_location
        )
        if not position_within_range(params.position, member_location_range):
            raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")
        node = node.referenced_declaration
    elif isinstance(node, (IdentifierPath, UserDefinedTypeName)):
        part = node.identifier_path_part_at(byte_offset)
        if part is None:
            raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")
        node = part.referenced_declaration
    elif isinstance(node, YulIdentifier):
        external_reference = node.external_reference
        if external_reference is None:
            raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")
        node = external_reference.referenced_declaration

    if not isinstance(node, DeclarationAbc):
        raise LspError(ErrorCodes.RequestFailed, "Cannot rename this symbol")

    return _generate_workspace_edit(node, params.new_name, context)


async def prepare_rename(
    context: LspContext,
    params: PrepareRenameParams,
) -> Union[Range, None]:
    """
    Prepares rename.
    """
    logger.debug(
        f"Requested prepare rename for file {params.text_document.uri} at {params.position}"
    )

    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    location = None
    if isinstance(node, Identifier):
        location = node.byte_location
        node = node.referenced_declaration
    elif isinstance(node, MemberAccess):
        location = node.member_byte_location
        node = node.referenced_declaration
    elif isinstance(node, (IdentifierPath, UserDefinedTypeName)):
        part = node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        location = part.byte_location
        node = part.referenced_declaration
    elif isinstance(node, YulIdentifier):
        external_reference = node.external_reference
        if external_reference is None:
            return None
        location = external_reference.identifier_byte_location
        node = external_reference.referenced_declaration
    elif isinstance(node, DeclarationAbc):
        name_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.name_location
        )
        if not position_within_range(params.position, name_location_range):
            return None
        location = node.name_location

    if not isinstance(node, DeclarationAbc) or location is None:
        return None
    return context.compiler.get_range_from_byte_offsets(path, location)



from __future__ import annotations

import re
from functools import lru_cache, partial
from typing import TYPE_CHECKING, FrozenSet, Iterator, List, Optional, Set, Tuple, Union

from ..abc import IrAbc
from ..meta.inheritance_specifier import InheritanceSpecifier
from ..meta.using_for_directive import UsingForDirective
from ..reference_resolver import CallbackParams
from .abc import DeclarationAbc

if TYPE_CHECKING:
    from woke.compiler import SolcOutputContractInfo
    from ..meta.source_unit import SourceUnit

from woke.ir.ast import (
    AstNodeId,
    SolcContractDefinition,
    SolcEnumDefinition,
    SolcErrorDefinition,
    SolcEventDefinition,
    SolcFunctionDefinition,
    SolcModifierDefinition,
    SolcStructDefinition,
    SolcStructuredDocumentation,
    SolcUserDefinedValueTypeDefinition,
    SolcUsingForDirective,
    SolcVariableDeclaration,
)
from woke.ir.declarations.enum_definition import EnumDefinition
from woke.ir.declarations.error_definition import ErrorDefinition
from woke.ir.declarations.event_definition import EventDefinition
from woke.ir.declarations.function_definition import FunctionDefinition
from woke.ir.declarations.modifier_definition import ModifierDefinition
from woke.ir.declarations.struct_definition import StructDefinition
from woke.ir.declarations.user_defined_value_type_definition import (
    UserDefinedValueTypeDefinition,
)
from woke.ir.declarations.variable_declaration import VariableDeclaration
from woke.ir.enums import ContractKind
from woke.ir.meta.structured_documentation import StructuredDocumentation
from woke.ir.utils import IrInitTuple


class ContractDefinition(DeclarationAbc):
    """
    Definition of a contract, library or interface. [byte_location][woke.ir.abc.IrAbc.byte_location] also includes the contract body.

    !!! example
        ```solidity
        contract C {
            uint x;
            function f() public {}
        }
        ```

        ```solidity
        interface I {
            function f() external;
        }
        ```

        ```solidity
        library L {
            function f() internal pure returns (uint) {
                return 7;
            }
        }
        ```
    """

    _ast_node: SolcContractDefinition
    _parent: SourceUnit

    _abstract: bool
    _base_contracts: List[InheritanceSpecifier]
    # ___dependencies
    _kind: ContractKind
    _fully_implemented: Optional[bool]
    _linearized_base_contracts: List[AstNodeId]
    # __scope
    _documentation: Optional[Union[StructuredDocumentation, str]]
    _compilation_info: Optional[SolcOutputContractInfo]
    _used_errors: List[AstNodeId]
    _enums: List[EnumDefinition]
    _errors: List[ErrorDefinition]
    _events: List[EventDefinition]
    _functions: List[FunctionDefinition]
    _modifiers: List[ModifierDefinition]
    _structs: List[StructDefinition]
    _user_defined_value_types: List[UserDefinedValueTypeDefinition]
    _using_for_directives: List[UsingForDirective]
    _declared_variables: List[VariableDeclaration]
    _used_events: List[AstNodeId]
    # _internal_function_ids

    _child_contracts: Set[ContractDefinition]

    def __init__(
        self, init: IrInitTuple, contract: SolcContractDefinition, parent: SourceUnit
    ):
        super().__init__(init, contract, parent)
        self._abstract = contract.abstract
        self._kind = contract.contract_kind
        self._fully_implemented = contract.fully_implemented
        self._linearized_base_contracts = list(contract.linearized_base_contracts)
        self._used_errors = (
            list(contract.used_errors) if contract.used_errors is not None else []
        )
        self._used_events = (
            list(contract.used_events) if contract.used_events is not None else []
        )

        if contract.documentation is None:
            self._documentation = None
        elif isinstance(contract.documentation, SolcStructuredDocumentation):
            self._documentation = StructuredDocumentation(
                init, contract.documentation, self
            )
        elif isinstance(contract.documentation, str):
            self._documentation = contract.documentation
        else:
            raise TypeError(
                f"Unknown type of documentation: {type(contract.documentation)}"
            )
        if init.contracts_info is not None and self.name in init.contracts_info:
            self._compilation_info = init.contracts_info[self.name]
        else:
            self._compilation_info = None

        self._base_contracts = []
        for base_contract in contract.base_contracts:
            self._base_contracts.append(InheritanceSpecifier(init, base_contract, self))
        self._child_contracts = set()

        self._enums = []
        self._errors = []
        self._events = []
        self._functions = []
        self._modifiers = []
        self._structs = []
        self._user_defined_value_types = []
        self._using_for_directives = []
        self._declared_variables = []

        for node in contract.nodes:
            if isinstance(node, SolcEnumDefinition):
                self._enums.append(EnumDefinition(init, node, self))
            elif isinstance(node, SolcErrorDefinition):
                self._errors.append(ErrorDefinition(init, node, self))
            elif isinstance(node, SolcEventDefinition):
                self._events.append(EventDefinition(init, node, self))
            elif isinstance(node, SolcFunctionDefinition):
                self._functions.append(FunctionDefinition(init, node, self))
            elif isinstance(node, SolcModifierDefinition):
                self._modifiers.append(ModifierDefinition(init, node, self))
            elif isinstance(node, SolcStructDefinition):
                self._structs.append(StructDefinition(init, node, self))
            elif isinstance(node, SolcUserDefinedValueTypeDefinition):
                self._user_defined_value_types.append(
                    UserDefinedValueTypeDefinition(init, node, self)
                )
            elif isinstance(node, SolcUsingForDirective):
                self._using_for_directives.append(UsingForDirective(init, node, self))
            elif isinstance(node, SolcVariableDeclaration):
                self._declared_variables.append(VariableDeclaration(init, node, self))

        init.reference_resolver.register_post_process_callback(self._post_process)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for base_contract in self._base_contracts:
            yield from base_contract
        if isinstance(self._documentation, StructuredDocumentation):
            yield from self._documentation
        for enum in self._enums:
            yield from enum
        for error in self._errors:
            yield from error
        for event in self._events:
            yield from event
        for function in self._functions:
            yield from function
        for modifier in self._modifiers:
            yield from modifier
        for struct in self._structs:
            yield from struct
        for user_defined_value_type in self._user_defined_value_types:
            yield from user_defined_value_type
        for using_for_directive in self._using_for_directives:
            yield from using_for_directive
        for declared_variable in self._declared_variables:
            yield from declared_variable

    def _post_process(self, callback_params: CallbackParams):
        base_contracts = []
        for base_contract in self._base_contracts:
            contract = base_contract.base_name.referenced_declaration
            assert isinstance(contract, ContractDefinition)
            contract._child_contracts.add(self)
            base_contracts.append(contract)

        for error in self.used_errors:
            error._used_in.append(self)

        self._reference_resolver.register_destroy_callback(
            self.file, partial(self._destroy, base_contracts)
        )

    def _destroy(self, base_contracts: List[ContractDefinition]) -> None:
        for base_contract in base_contracts:
            base_contract._child_contracts.remove(self)
        for error in self.used_errors:
            error._used_in.remove(self)

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        CONTRACT_RE = re.compile(
            r"^\s*(abstract\s)?\s*contract\s+(?P<name>{identifier})".format(
                identifier=IDENTIFIER
            ).encode("utf-8")
        )
        INTERFACE_RE = re.compile(
            r"^\s*interface\s+(?P<name>{identifier})".format(
                identifier=IDENTIFIER
            ).encode("utf-8")
        )
        LIBRARY_RE = re.compile(
            r"^\s*library\s+(?P<name>{identifier})".format(
                identifier=IDENTIFIER
            ).encode("utf-8")
        )

        if self.kind == ContractKind.CONTRACT:
            match = CONTRACT_RE.match(self._source)
        elif self.kind == ContractKind.INTERFACE:
            match = INTERFACE_RE.match(self._source)
        elif self.kind == ContractKind.LIBRARY:
            match = LIBRARY_RE.match(self._source)
        else:
            raise ValueError(f"Unknown contract kind: {self.kind}")
        assert match

        return self.byte_location[0] + match.start("name"), self.byte_location[
            0
        ] + match.end("name")

    @property
    def parent(self) -> SourceUnit:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return self._name

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        ret = f"{'abstract ' if self.abstract else ''}{self.kind} {self.name}"
        ret += (
            " is " + ", ".join(spec.source for spec in self.base_contracts)
            if len(self.base_contracts) > 0
            else ""
        )
        if isinstance(self.documentation, StructuredDocumentation):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.text.splitlines())
                + "\n"
                + ret
            )
        elif isinstance(self.documentation, str):
            return (
                "/// "
                + "\n///".join(line for line in self.documentation.splitlines())
                + "\n"
                + ret
            )
        else:
            return ret

    @property
    def abstract(self) -> bool:
        """
        Is `False` if the [kind][woke.ir.declarations.contract_definition.ContractDefinition.kind] is [ContractKind.LIBRARY][woke.ir.enums.ContractKind.LIBRARY] or [ContractKind.INTERFACE][woke.ir.enums.ContractKind.INTERFACE].

        Returns:
            `True` if the contract is abstract, `False` otherwise.
        """
        return self._abstract

    @property
    def base_contracts(self) -> Tuple[InheritanceSpecifier, ...]:
        """
        Returns base contracts as specified in the source code. Does not return all base contracts (recursively).
        !!! example
            `A1` lists the interface `I` as a base contract.

            `A2` lists the interface `I` as a base contract.

            `B` lists the contracts `A1` and `A2` as base contracts.
            ```solidity
            interface I {}
            contract A1 is I {}
            contract A2 is I {}
            contract B is A1, A2 {}
            ```
        Returns:
            Base contracts of this contract.
        """
        return tuple(self._base_contracts)

    @property
    def child_contracts(self) -> FrozenSet[ContractDefinition]:
        """
        Returns:
            Contracts that list this contract in their [base_contracts][woke.ir.declarations.contract_definition.ContractDefinition.base_contracts] property.
        """
        return frozenset(self._child_contracts)

    @property
    def kind(self) -> ContractKind:
        """
        Returns:
            Contract kind.
        """
        return self._kind

    @property
    def fully_implemented(self) -> Optional[bool]:
        """
        Is `None` when a file that imports this contract cannot be compiled. This may happen in the LSP server where partial project analysis is supported.
        Returns:
            `True` if all functions and modifiers of the contract are implemented, `False` otherwise.
        """
        return self._fully_implemented

    @property
    def linearized_base_contracts(self) -> Tuple[ContractDefinition, ...]:
        """
        Returns:
            C3 linearized list of all base contracts.
        """
        base_contracts = []
        for base_contract in self._linearized_base_contracts:
            node = self._reference_resolver.resolve_node(base_contract, self._cu_hash)
            assert isinstance(node, ContractDefinition)
            base_contracts.append(node)
        return tuple(base_contracts)

    @property
    def used_errors(self) -> Tuple[ErrorDefinition, ...]:
        """
        Returns:
            Errors used in the contract.
        """
        used_errors = []
        for error in self._used_errors:
            node = self._reference_resolver.resolve_node(error, self._cu_hash)
            assert isinstance(node, ErrorDefinition)
            used_errors.append(node)
        return tuple(used_errors)

    @property
    def used_events(self) -> Tuple[EventDefinition, ...]:
        """
        Returns:
            Events emitted by the contract as well as all events defined and inherited by the contract.
        """
        used_events = []
        for event in self._used_events:
            node = self._reference_resolver.resolve_node(event, self._cu_hash)
            assert isinstance(node, EventDefinition)
            used_events.append(node)
        return tuple(used_events)

    @property
    def documentation(self) -> Optional[Union[StructuredDocumentation, str]]:
        """
        Of [StructuredDocumentation][woke.ir.meta.structured_documentation.StructuredDocumentation] type since Solidity 0.6.3.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation of this contract, if any.
        """
        return self._documentation

    @property
    def compilation_info(self) -> Optional[SolcOutputContractInfo]:
        return self._compilation_info

    @property
    def enums(self) -> Tuple[EnumDefinition, ...]:
        """
        Returns:
            Enum definitions contained in this contract.
        """
        return tuple(self._enums)

    @property
    def errors(self) -> Tuple[ErrorDefinition, ...]:
        """
        Returns:
            Error definitions contained in this contract.
        """
        return tuple(self._errors)

    @property
    def events(self) -> Tuple[EventDefinition, ...]:
        """
        Returns:
            Event definitions contained in this contract.
        """
        return tuple(self._events)

    @property
    def functions(self) -> Tuple[FunctionDefinition, ...]:
        """
        Returns:
            Function definitions contained in this contract.
        """
        return tuple(self._functions)

    @property
    def modifiers(self) -> Tuple[ModifierDefinition, ...]:
        """
        Returns:
            Modifier definitions contained in this contract.
        """
        return tuple(self._modifiers)

    @property
    def structs(self) -> Tuple[StructDefinition, ...]:
        """
        Returns:
            Struct definitions contained in this contract.
        """
        return tuple(self._structs)

    @property
    def user_defined_value_types(self) -> Tuple[UserDefinedValueTypeDefinition, ...]:
        """
        Returns:
            User defined value type definitions contained in this contract.
        """
        return tuple(self._user_defined_value_types)

    @property
    def using_for_directives(self) -> Tuple[UsingForDirective, ...]:
        """
        Returns:
            Using for directives contained in this contract.
        """
        return tuple(self._using_for_directives)

    @property
    def declared_variables(self) -> Tuple[VariableDeclaration, ...]:
        """
        Returns:
            Variable declarations contained in this contract.
        """
        return tuple(self._declared_variables)

    def declarations_iter(self) -> Iterator[DeclarationAbc]:
        """
        Yields:
            All declarations contained in this contract.
        """
        yield from self.enums
        for enum in self.enums:
            yield from enum.values
        yield from self.errors
        yield from self.events
        yield from self.functions
        yield from self.modifiers
        yield from self.structs
        yield from self.user_defined_value_types
        yield from self.declared_variables



import asyncio
import logging
import traceback
import uuid
from copy import deepcopy
from pathlib import Path
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from pydantic.error_wrappers import ValidationError

from woke.utils import StrEnum

from ..config import WokeConfig
from .commands import (
    generate_cfg_handler,
    generate_imports_graph_handler,
    generate_inheritance_graph_handler,
    generate_linearized_inheritance_graph_handler,
)
from .common_structures import (
    ConfigurationItem,
    ConfigurationParams,
    CreateFilesParams,
    DeleteFilesParams,
    DidChangeConfigurationParams,
    DocumentFilter,
    DocumentUri,
    ExecuteCommandOptions,
    ExecuteCommandParams,
    InitializedParams,
    InitializeError,
    InitializeParams,
    LogMessageParams,
    LogTraceParams,
    MessageType,
    PositionEncodingKind,
    ProgressParams,
    RenameFilesParams,
    SetTraceParams,
    ShowMessageParams,
    WorkDoneProgressBegin,
    WorkDoneProgressCreateParams,
    WorkDoneProgressEnd,
    WorkDoneProgressReport,
)
from .context import LspContext
from .document_sync import (
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    TextDocumentSyncKind,
    TextDocumentSyncOptions,
    WillSaveTextDocumentParams,
)
from .exceptions import LspError
from .features.code_action import CodeActionOptions, CodeActionParams, code_action
from .features.code_lens import CodeLensOptions, CodeLensParams, code_lens
from .features.completion import CompletionOptions, CompletionParams, completion
from .features.definition import DefinitionParams, definition
from .features.document_link import (
    DocumentLinkOptions,
    DocumentLinkParams,
    document_link,
)
from .features.document_symbol import DocumentSymbolParams, document_symbol
from .features.hover import HoverParams, hover
from .features.implementation import ImplementationParams, implementation
from .features.references import ReferenceParams, references
from .features.rename import (
    PrepareRenameParams,
    RenameOptions,
    RenameParams,
    prepare_rename,
    rename,
)
from .features.type_definition import TypeDefinitionParams, type_definition
from .features.type_hierarchy import (
    TypeHierarchyPrepareParams,
    TypeHierarchySubtypesParams,
    TypeHierarchySupertypesParams,
    prepare_type_hierarchy,
    subtypes,
    supertypes,
)
from .lsp_data_model import LspModel
from .methods import RequestMethodEnum
from .protocol_structures import (
    CancelParams,
    ErrorCodes,
    NotificationMessage,
    RequestMessage,
    ResponseError,
    ResponseMessage,
)
from .rpc_protocol import RpcProtocol
from .server_capabilities import (
    FileOperationFilter,
    FileOperationPattern,
    FileOperationPatternKind,
    FileOperationRegistrationOptions,
    InitializeResult,
    ServerCapabilities,
    ServerCapabilitiesWorkspace,
    ServerCapabilitiesWorkspaceFileOperations,
    WorkspaceFoldersServerCapabilities,
)
from .utils.uri import uri_to_path

logger = logging.getLogger(__name__)

ConfigPath = Tuple[Union[str, int], ...]


class CommandsEnum(StrEnum):
    GENERATE_CFG = "woke.generate.control_flow_graph"
    GENERATE_IMPORTS_GRAPH = "woke.generate.imports_graph"
    GENERATE_INHERITANCE_GRAPH = "woke.generate.inheritance_graph"
    GENERATE_INHERITANCE_GRAPH_FULL = "woke.generate.inheritance_graph_full"
    GENERATE_LINEARIZED_INHERITANCE_GRAPH = "woke.generate.linearized_inheritance_graph"
    LSP_FORCE_RECOMPILE = "woke.lsp.force_recompile"
    LSP_FORCE_RERUN_DETECTORS = "woke.lsp.force_rerun_detectors"


class LspServer:
    __initialized: bool
    __cli_config: WokeConfig
    __workspaces: Dict[Path, LspContext]
    __user_config: Optional[WokeConfig]
    __main_workspace: Optional[LspContext]
    __workspace_path: Optional[Path]
    __protocol: RpcProtocol
    __run: bool
    __request_id_counter: int
    __sent_requests: Dict[Union[int, str], asyncio.Event]
    __message_responses: Dict[Union[int, str], ResponseMessage]
    __running_tasks: Set[asyncio.Task]
    __request_tasks: Dict[Union[int, str], asyncio.Task]

    __method_mapping: Dict[str, Tuple[Callable, Optional[Type[LspModel]]]]
    __notification_mapping: Dict[str, Tuple[Callable, Optional[Type[LspModel]]]]

    def __init__(
        self,
        config: WokeConfig,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        self.__initialized = False
        self.__cli_config = config
        self.__workspaces = {}
        self.__user_config = None
        self.__main_workspace = None
        self.__workspace_path = None
        self.__protocol = RpcProtocol(reader, writer)
        self.__run = True
        self.__request_id_counter = 0
        self.__sent_requests = {}
        self.__message_responses = {}
        self.__running_tasks = set()
        self.__request_tasks = {}

        self.__method_mapping = {
            RequestMethodEnum.INITIALIZE: (self._initialize, InitializeParams),
            RequestMethodEnum.SHUTDOWN: (self._shutdown, None),
            RequestMethodEnum.DOCUMENT_LINK: (
                self._workspace_route,
                DocumentLinkParams,
            ),
            RequestMethodEnum.PREPARE_TYPE_HIERARCHY: (
                self._workspace_route,
                TypeHierarchyPrepareParams,
            ),
            RequestMethodEnum.TYPE_HIERARCHY_SUPERTYPES: (
                self._workspace_route,
                TypeHierarchySupertypesParams,
            ),
            RequestMethodEnum.TYPE_HIERARCHY_SUBTYPES: (
                self._workspace_route,
                TypeHierarchySubtypesParams,
            ),
            RequestMethodEnum.REFERENCES: (self._workspace_route, ReferenceParams),
            RequestMethodEnum.DOCUMENT_SYMBOL: (
                self._workspace_route,
                DocumentSymbolParams,
            ),
            RequestMethodEnum.DEFINITION: (self._workspace_route, DefinitionParams),
            RequestMethodEnum.TYPE_DEFINITION: (
                self._workspace_route,
                TypeDefinitionParams,
            ),
            RequestMethodEnum.IMPLEMENTATION: (
                self._workspace_route,
                ImplementationParams,
            ),
            RequestMethodEnum.CODE_LENS: (self._workspace_route, CodeLensParams),
            RequestMethodEnum.PREPARE_RENAME: (
                self._workspace_route,
                PrepareRenameParams,
            ),
            RequestMethodEnum.RENAME: (self._workspace_route, RenameParams),
            RequestMethodEnum.WORKSPACE_EXECUTE_COMMAND: (
                self._workspace_execute_command,
                ExecuteCommandParams,
            ),
            RequestMethodEnum.HOVER: (self._workspace_route, HoverParams),
            RequestMethodEnum.COMPLETION: (self._workspace_route, CompletionParams),
            RequestMethodEnum.CODE_ACTION: (self._workspace_route, CodeActionParams),
        }

        self.__notification_mapping = {
            RequestMethodEnum.INITIALIZED: (self._initialized, InitializedParams),
            RequestMethodEnum.CANCEL_REQUEST: (self._cancel_request, CancelParams),
            RequestMethodEnum.PROGRESS: (self._progress, ProgressParams),
            RequestMethodEnum.LOG_TRACE: (self._log_trace, LogTraceParams),
            RequestMethodEnum.SET_TRACE: (self._set_trace, SetTraceParams),
            RequestMethodEnum.TEXT_DOCUMENT_DID_OPEN: (
                self._text_document_did_open,
                DidOpenTextDocumentParams,
            ),
            RequestMethodEnum.TEXT_DOCUMENT_DID_CHANGE: (
                self._workspace_route,
                DidChangeTextDocumentParams,
            ),
            RequestMethodEnum.TEXT_DOCUMENT_WILL_SAVE: (
                self._workspace_route,
                WillSaveTextDocumentParams,
            ),
            RequestMethodEnum.TEXT_DOCUMENT_DID_SAVE: (
                self._workspace_route,
                DidSaveTextDocumentParams,
            ),
            RequestMethodEnum.TEXT_DOCUMENT_DID_CLOSE: (
                self._workspace_route,
                DidCloseTextDocumentParams,
            ),
            RequestMethodEnum.WORKSPACE_DID_CHANGE_CONFIGURATION: (
                self._workspace_did_change_configuration,
                DidChangeConfigurationParams,
            ),
            RequestMethodEnum.WORKSPACE_DID_CREATE_FILES: (
                self._workspace_did_create_files,
                CreateFilesParams,
            ),
            RequestMethodEnum.WORKSPACE_DID_RENAME_FILES: (
                self._workspace_did_rename_files,
                RenameFilesParams,
            ),
            RequestMethodEnum.WORKSPACE_DID_DELETE_FILES: (
                self._workspace_did_delete_files,
                DeleteFilesParams,
            ),
        }

    def _task_done_callback(self, task: asyncio.Task) -> None:
        try:
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:

            def _callback(task: asyncio.Task) -> None:
                for t in self.__running_tasks:
                    t.cancel()

            logger.exception(e)
            try:
                t = asyncio.create_task(
                    self.log_message(traceback.format_exc(), MessageType.ERROR)
                )
                t.add_done_callback(_callback)
            except Exception as e:
                logger.exception(e)

                for task in self.__running_tasks:
                    task.cancel()
        finally:
            self.__running_tasks.remove(task)

    def create_task(self, coroutine: Coroutine) -> asyncio.Task:
        task = asyncio.create_task(coroutine)
        self.__running_tasks.add(task)
        task.add_done_callback(self._task_done_callback)
        return task

    def create_request_task(
        self, coroutine: Coroutine, request_id: Union[int, str]
    ) -> asyncio.Task:
        def _callback(task: asyncio.Task) -> None:
            if request_id in self.__request_tasks:
                del self.__request_tasks[request_id]

        task = self.create_task(coroutine)
        self.__request_tasks[request_id] = task
        task.add_done_callback(_callback)
        return task

    async def run(self) -> None:
        task = self.create_task(self._main_task())
        try:
            await task
        except asyncio.CancelledError:
            pass

    async def _main_task(self) -> None:
        messages_queue = asyncio.Queue()
        self.create_task(self._messages_loop(messages_queue))

        try:
            while self.__run:
                message = await self.__protocol.receive()
                if isinstance(message, ResponseMessage):
                    await self._handle_response(message)
                else:
                    await messages_queue.put(message)
        except ConnectionError:
            pass

        for task in self.__running_tasks:
            task.cancel()

    async def _messages_loop(self, queue: asyncio.Queue) -> NoReturn:
        while True:
            message = await queue.get()
            if isinstance(message, RequestMessage):
                self.create_request_task(self._handle_message(message), message.id)
            elif isinstance(message, NotificationMessage):
                if message.method != RequestMethodEnum.INITIALIZED:
                    self.create_task(self._handle_notification(message))
                else:
                    await self._handle_notification(message)
            else:
                raise Exception("Unknown message type")

    async def send_request(self, method: RequestMethodEnum, params: Any = None) -> Any:
        request = RequestMessage(
            jsonrpc="2.0", id=self.__request_id_counter, method=method, params=params
        )
        self.__sent_requests[request.id] = asyncio.Event()
        self.__request_id_counter += 1

        logger.debug(f"Sending request:\n{request}")
        await self.__protocol.send(request)
        await self.__sent_requests[request.id].wait()
        self.__sent_requests.pop(request.id)
        response = self.__message_responses.pop(request.id)

        if response.error is not None:
            raise LspError(
                response.error.code, response.error.message, response.error.data
            )
        return response.result

    async def send_notification(
        self, method: str, params: Optional[Any] = None
    ) -> None:
        notification = NotificationMessage(
            jsonrpc="2.0",
            method=method,
            params=params,
        )
        logger.debug(f"Sending notification:\n{notification}")
        await self.__protocol.send(notification)

    async def log_message(self, message: str, type: MessageType) -> None:
        params = LogMessageParams(
            type=type,
            message=message,
        )
        await self.send_notification(RequestMethodEnum.WINDOW_LOG_MESSAGE, params)

    async def show_message(self, message: str, type: MessageType) -> None:
        params = ShowMessageParams(
            type=type,
            message=message,
        )
        await self.send_notification(RequestMethodEnum.WINDOW_SHOW_MESSAGE, params)

    async def progress_begin(
        self,
        title: str,
        message: Optional[str] = None,
        percentage: Optional[int] = None,
        cancellable: Optional[bool] = None,
    ) -> Optional[str]:
        token = str(uuid.uuid4())
        params = WorkDoneProgressCreateParams(token=token)
        try:
            await self.send_request(
                RequestMethodEnum.WINDOW_WORK_DONE_PROGRESS_CREATE, params
            )
        except LspError:
            return None

        params = ProgressParams(
            token=token,
            value=WorkDoneProgressBegin(
                kind="begin",
                title=title,
                message=message,
                percentage=percentage,
                cancellable=cancellable,
            ),
        )
        await self.send_notification(RequestMethodEnum.PROGRESS, params)
        return token

    async def progress_report(
        self,
        token: str,
        message: Optional[str] = None,
        percentage: Optional[int] = None,
        cancellable: Optional[bool] = None,
    ) -> None:
        params = ProgressParams(
            token=token,
            value=WorkDoneProgressReport(
                kind="report",
                message=message,
                percentage=percentage,
                cancellable=cancellable,
            ),
        )
        await self.send_notification(RequestMethodEnum.PROGRESS, params)

    async def progress_end(self, token: str, message: Optional[str] = None) -> None:
        params = ProgressParams(
            token=token,
            value=WorkDoneProgressEnd(
                kind="end",
                message=message,
            ),
        )
        await self.send_notification(RequestMethodEnum.PROGRESS, params)

    async def _handle_message(self, request: RequestMessage) -> None:
        logger.info(f"Message received: {request}")

        # Init before request needed
        if request.method != RequestMethodEnum.INITIALIZE and not self.__initialized:
            response = self._serve_error(
                request,
                ErrorCodes.ServerNotInitialized,
                "Server has not been initialized",
            )
            await self.__protocol.send(response)
            return

        # Handling request
        try:
            response = await self._serve_response(request)
        except LspError as e:
            response = self._serve_error(request, e.code, e.message)
        await self.__protocol.send(response)

    async def _handle_notification(self, notification: NotificationMessage) -> None:
        logger.info(f"Notification received: {notification}")

        if not self.__initialized and notification.method != RequestMethodEnum.EXIT:
            return

        try:
            n, params_type = self.__notification_mapping[notification.method]
        except KeyError:
            logger.error(
                f"Incoming notification type '{notification.method}' not implemented."
            )
            raise NotImplementedError()

        if params_type is not None:
            await n(params_type.parse_obj(notification.params))
        else:
            await n(None)

    async def _handle_response(self, response: ResponseMessage) -> None:
        logger.info(f"Response received: {response}")

        if response.id is None:
            logger.error(f"Response without id: {response}")
            return

        try:
            self.__message_responses[response.id] = response
            self.__sent_requests[response.id].set()
        except KeyError:
            logger.error(
                f"Received response with id {response.id} but no such request was sent."
            )

    async def _serve_response(self, request: RequestMessage) -> ResponseMessage:
        try:
            m, params_type = self.__method_mapping[request.method]
        except KeyError:
            logger.error(f"Incoming method type '{request.method}' not implemented.")
            raise NotImplementedError()

        if params_type is not None:
            response = await m(params_type.parse_obj(request.params))
        else:
            response = await m(None)

        response_message = ResponseMessage(
            jsonrpc="2.0", id=request.id, result=response, error=None
        )
        logger.info(f"Serving response: {response_message}")
        return response_message

    @staticmethod
    def _serve_error(
        request: RequestMessage, error_code: int, msg: str
    ) -> ResponseMessage:
        response_error = ResponseError(code=error_code, message=msg, data=None)
        response_message = ResponseMessage(
            jsonrpc="2.0", id=request.id, error=response_error, result=None
        )
        logger.warning(f"Serving error response: {response_message}")
        return response_message

    async def _initialize(self, params: InitializeParams) -> InitializeResult:
        if self.__initialized:
            raise LspError(ErrorCodes.InvalidRequest, "Server already initialized")

        if params.workspace_folders is not None:
            if len(params.workspace_folders) != 1:
                raise LspError(
                    ErrorCodes.RequestFailed,
                    "Multi-root workspaces are not supported.",
                    InitializeError(retry=False),
                )
            path = uri_to_path(params.workspace_folders[0].uri).resolve(strict=True)
        elif params.root_uri is not None:
            path = uri_to_path(params.root_uri).resolve(strict=True)
        elif params.root_path is not None:
            path = Path(params.root_path).resolve(strict=True)
        else:
            path = None

        self.__initialized = True
        self.__workspace_path = path

        solidity_registration = FileOperationRegistrationOptions(
            filters=[
                FileOperationFilter(
                    pattern=FileOperationPattern(
                        glob="**/*.sol",
                        matches=FileOperationPatternKind.FILE,
                    )
                )
            ]
        )

        server_capabilities = ServerCapabilities(
            position_encoding=PositionEncodingKind.UTF16,
            text_document_sync=TextDocumentSyncOptions(
                open_close=True, change=TextDocumentSyncKind.INCREMENTAL
            ),
            document_link_provider=DocumentLinkOptions(
                resolve_provider=False,
            ),
            type_hierarchy_provider=True,
            references_provider=True,
            document_symbol_provider=True,
            workspace=ServerCapabilitiesWorkspace(
                workspace_folders=WorkspaceFoldersServerCapabilities(
                    supported=False,
                ),
                file_operations=ServerCapabilitiesWorkspaceFileOperations(
                    did_create=solidity_registration,
                    did_rename=solidity_registration,
                    did_delete=solidity_registration,
                ),
            ),
            definition_provider=True,
            type_definition_provider=True,
            implementation_provider=True,
            code_lens_provider=CodeLensOptions(
                resolve_provider=False,
            ),
            rename_provider=RenameOptions(
                prepare_provider=True,
            ),
            execute_command_provider=ExecuteCommandOptions(
                commands=[command for command in CommandsEnum]
            ),
            hover_provider=True,
            completion_provider=CompletionOptions(
                trigger_characters=[".", '"', "'"],
                all_commit_characters=None,
                resolve_provider=None,
                completion_item=None,
            ),
            code_action_provider=CodeActionOptions(
                code_action_kinds=None,
                resolve_provider=False,
            ),
        )
        return InitializeResult(capabilities=server_capabilities, server_info=None)

    async def _cancel_request(self, params: CancelParams) -> None:
        if params.id in self.__request_tasks:
            self.__request_tasks[params.id].cancel()
            del self.__request_tasks[params.id]

    async def _progress(self, params: ProgressParams) -> None:
        pass

    async def _log_trace(self, params: LogTraceParams) -> None:
        pass

    async def _set_trace(self, params: SetTraceParams) -> None:
        pass

    async def _shutdown(self, params: Any) -> None:
        self.__run = False

    async def _parse_config(
        self, raw_config: dict, workspace_path: Path
    ) -> Tuple[dict, Set[ConfigPath], Set[ConfigPath]]:
        removed_options: Set[ConfigPath] = set()

        def _normalize_config(config: Union[dict, list], config_path: ConfigPath):
            if isinstance(config, dict):
                for k in list(config):
                    v = config[k]
                    if isinstance(v, (dict, list)):
                        if len(v) == 0:
                            del config[k]
                            removed_options.add(config_path + (k,))
                        else:
                            _normalize_config(v, config_path + (k,))
                    elif isinstance(v, str) and len(v.strip()) == 0:
                        del config[k]
                        removed_options.add(config_path + (k,))
            else:
                for no, item in enumerate(config):
                    if isinstance(item, (dict, list)):
                        if len(item) == 0:
                            config.remove(item)
                            removed_options.add(config_path + (no,))
                        else:
                            _normalize_config(item, config_path + (no,))
                    elif isinstance(item, str) and len(item.strip()) == 0:
                        config.remove(item)
                        removed_options.add(config_path + (no,))

        _normalize_config(raw_config, tuple())

        run = True
        invalid_options: Set[ConfigPath] = set()
        while run:
            try:
                WokeConfig.fromdict(
                    raw_config,
                    project_root_path=workspace_path,
                )
                run = False
            except ValidationError as e:
                to_be_removed = []
                for error in e.errors():
                    invalid_option = raw_config
                    for segment in error["loc"][:-1]:
                        invalid_option = invalid_option[segment]

                    if isinstance(invalid_option, list):
                        val = invalid_option[error["loc"][-1]]  # type: ignore
                        invalid_options.add(error["loc"][:-1] + (val,))
                        to_be_removed.append((error["loc"][:-1], val))
                    elif isinstance(invalid_option, dict):
                        invalid_options.add(error["loc"])
                        to_be_removed.append((error["loc"][:-1], error["loc"][-1]))
                    else:
                        raise NotImplementedError()

                for p, val in to_be_removed:
                    path = raw_config
                    for segment in p:
                        path = path[segment]

                    if isinstance(path, list):
                        path.remove(val)
                    else:
                        del path[val]
        if len(invalid_options) > 0:
            message = (
                "Failed to parse the following config options, using defaults:\n"
                + "\n".join(
                    f"    woke -> {' -> '.join(str(segment) for segment in option)}"
                    for option in invalid_options
                )
            )
            await self.log_message(message, MessageType.WARNING)

        return raw_config, invalid_options, removed_options

    async def _create_config(self, workspace_path: Path) -> WokeConfig:
        code_config = await self.get_configuration()
        assert isinstance(code_config, list)
        assert len(code_config) == 1
        assert isinstance(code_config[0], dict)
        raw_config = code_config[0]

        raw_config, _, _ = await self._parse_config(raw_config, workspace_path)

        return WokeConfig.fromdict(
            raw_config,
            project_root_path=workspace_path,
        )

    async def _handle_config_change(self, raw_config: dict) -> None:
        def key_in_nested_dict(key: Tuple, d: Dict) -> bool:
            try:
                for k in key:
                    d = d[k]
                return True
            except KeyError:
                return False

        for context in self.__workspaces.values():
            raw_config_copy = deepcopy(raw_config)
            (
                raw_config_copy,
                invalid_options,
                removed_options,
            ) = await self._parse_config(
                raw_config_copy, context.config.project_root_path
            )

            changed = context.config.update(
                raw_config_copy, invalid_options.union(removed_options)
            )

            if key_in_nested_dict(("compiler", "solc"), changed):
                await context.compiler.force_recompile()
            if key_in_nested_dict(("lsp", "detectors"), changed) or key_in_nested_dict(
                ("detectors",), changed
            ):
                await context.compiler.force_rerun_detectors()
            if key_in_nested_dict(("lsp", "code_lens"), changed):
                try:
                    await self.send_request(
                        RequestMethodEnum.WORKSPACE_CODE_LENS_REFRESH, None
                    )
                except LspError:
                    pass

    async def _initialized(self, params: InitializedParams) -> None:
        if self.__workspace_path is not None:
            self.__main_workspace = LspContext(
                self, await self._create_config(self.__workspace_path), True
            )
            self.__workspaces[self.__workspace_path] = self.__main_workspace
            self.__main_workspace.run()

    async def _workspace_did_change_configuration(
        self, params: DidChangeConfigurationParams
    ) -> None:
        logger.debug(f"Received configuration change: {params}")
        if "woke" in params.settings:
            await self._handle_config_change(params.settings["woke"])

    async def _get_workspace(self, uri: DocumentUri) -> LspContext:
        path = uri_to_path(uri)
        matching_workspaces = []
        for workspace in self.__workspaces.values():
            try:
                matching_workspaces.append(
                    (workspace, path.relative_to(workspace.config.project_root_path))
                )
            except ValueError:
                pass

        if len(matching_workspaces) == 0:
            context = LspContext(self, await self._create_config(path.parent), False)
            self.__workspaces[path.parent] = context
            context.run()
            return context
        else:
            return min(matching_workspaces, key=lambda x: len(x[1].parts))[0]

    async def _workspace_route(self, params: Any) -> Any:
        if isinstance(
            params, (TypeHierarchySupertypesParams, TypeHierarchySubtypesParams)
        ):
            uri = params.item.data.uri
        else:
            uri = params.text_document.uri

        context = await self._get_workspace(uri)

        if isinstance(params, DocumentLinkParams):
            return await document_link(context, params)
        elif isinstance(params, TypeHierarchyPrepareParams):
            return await prepare_type_hierarchy(context, params)
        elif isinstance(params, TypeHierarchySupertypesParams):
            return await supertypes(context, params)
        elif isinstance(params, TypeHierarchySubtypesParams):
            return await subtypes(context, params)
        elif isinstance(params, ReferenceParams):
            return await references(context, params)
        elif isinstance(params, DocumentSymbolParams):
            return await document_symbol(context, params)
        elif isinstance(params, DefinitionParams):
            return await definition(context, params)
        elif isinstance(params, TypeDefinitionParams):
            return await type_definition(context, params)
        elif isinstance(params, ImplementationParams):
            return await implementation(context, params)
        elif isinstance(params, CodeLensParams):
            return await code_lens(context, params)
        elif isinstance(params, PrepareRenameParams):
            return await prepare_rename(context, params)
        elif isinstance(params, RenameParams):
            return await rename(context, params)
        elif isinstance(params, DidChangeTextDocumentParams):
            return await self._text_document_did_change(context, params)
        elif isinstance(params, WillSaveTextDocumentParams):
            return await self._text_document_will_save(context, params)
        elif isinstance(params, DidSaveTextDocumentParams):
            return await self._text_document_did_save(context, params)
        elif isinstance(params, DidCloseTextDocumentParams):
            return await self._text_document_did_close(context, params)
        elif isinstance(params, HoverParams):
            return await hover(context, params)
        elif isinstance(params, CompletionParams):
            return await completion(context, params)
        elif isinstance(params, CodeActionParams):
            return await code_action(context, params)
        else:
            raise NotImplementedError(f"Unhandled request: {type(params)}")

    async def _text_document_did_open(self, params: DidOpenTextDocumentParams) -> None:
        path = uri_to_path(params.text_document.uri)
        matching_workspaces = []
        for workspace in self.__workspaces.values():
            try:
                matching_workspaces.append(
                    (workspace, path.relative_to(workspace.config.project_root_path))
                )
            except ValueError:
                pass

        if len(matching_workspaces) == 0:
            context = LspContext(self, await self._create_config(path.parent), False)
            self.__workspaces[path.parent] = context
            context.run()
        else:
            context = min(matching_workspaces, key=lambda x: len(x[1].parts))[0]

        await context.compiler.add_change(params)
        await context.parser.add_change(params)

    @staticmethod
    async def _text_document_did_change(
        context: LspContext, params: DidChangeTextDocumentParams
    ) -> None:
        await context.compiler.add_change(params)
        await context.parser.add_change(params)

    @staticmethod
    async def _text_document_will_save(
        context: LspContext, params: WillSaveTextDocumentParams
    ) -> None:
        pass

    @staticmethod
    async def _text_document_did_save(
        context: LspContext, params: DidSaveTextDocumentParams
    ) -> None:
        pass

    async def _text_document_did_close(
        self, context: LspContext, params: DidCloseTextDocumentParams
    ) -> None:
        await context.compiler.add_change(params)
        await context.parser.add_change(params)

        if context != self.__main_workspace:
            path = uri_to_path(params.text_document.uri)
            await context.diagnostics_queue.put((path, set()))
        # TODO: remove the workspace from the dict of workspaces (if not main workspace and no other files are open)

    async def _workspace_did_create_files(self, params: CreateFilesParams) -> None:
        assert self.__main_workspace is not None
        await self.__main_workspace.compiler.add_change(params)
        await self.__main_workspace.parser.add_change(params)

    async def _workspace_did_rename_files(self, params: RenameFilesParams) -> None:
        assert self.__main_workspace is not None
        await self.__main_workspace.compiler.add_change(params)
        await self.__main_workspace.parser.add_change(params)

    async def _workspace_did_delete_files(self, params: DeleteFilesParams) -> None:
        assert self.__main_workspace is not None
        await self.__main_workspace.compiler.add_change(params)
        await self.__main_workspace.parser.add_change(params)

    async def _workspace_execute_command(
        self, params: ExecuteCommandParams
    ) -> Optional[Any]:
        command = params.command
        if command == CommandsEnum.LSP_FORCE_RECOMPILE:
            for context in self.__workspaces.values():
                await context.compiler.force_recompile()
            return None
        elif command == CommandsEnum.LSP_FORCE_RERUN_DETECTORS:
            for context in self.__workspaces.values():
                await context.compiler.force_rerun_detectors()
            return None
        elif command == CommandsEnum.GENERATE_CFG:
            if params.arguments is None or len(params.arguments) != 2:
                raise LspError(
                    ErrorCodes.InvalidParams,
                    f"Expected 2 arguments for `{CommandsEnum.GENERATE_CFG}` command",
                )
            document_uri = DocumentUri(params.arguments[0])
            canonical_name = str(params.arguments[1])
            context = await self._get_workspace(document_uri)
            return await generate_cfg_handler(context, document_uri, canonical_name)
        elif command == CommandsEnum.GENERATE_IMPORTS_GRAPH:
            if params.arguments is None or len(params.arguments) == 0:
                if self.__main_workspace is None:
                    raise LspError(ErrorCodes.RequestFailed, "No workspace open")
                return await generate_imports_graph_handler(self.__main_workspace)
            else:
                raise LspError(
                    ErrorCodes.InvalidParams,
                    f"Expected 0 arguments for `{CommandsEnum.GENERATE_IMPORTS_GRAPH}` command",
                )
        elif command == CommandsEnum.GENERATE_INHERITANCE_GRAPH:
            if params.arguments is not None and len(params.arguments) == 2:
                document_uri = DocumentUri(params.arguments[0])
                canonical_name = str(params.arguments[1])
                context = await self._get_workspace(document_uri)
                return await generate_inheritance_graph_handler(
                    context, (document_uri, canonical_name)
                )
            else:
                raise LspError(
                    ErrorCodes.InvalidParams,
                    f"Expected 0 or 2 arguments for `{CommandsEnum.GENERATE_INHERITANCE_GRAPH}` command",
                )
        elif command == CommandsEnum.GENERATE_INHERITANCE_GRAPH_FULL:
            if params.arguments is None or len(params.arguments) == 0:
                if self.__main_workspace is None:
                    raise LspError(ErrorCodes.RequestFailed, "No workspace open")
                return await generate_inheritance_graph_handler(
                    self.__main_workspace, None
                )
            else:
                raise LspError(
                    ErrorCodes.InvalidParams,
                    f"Expected 0 arguments for `{CommandsEnum.GENERATE_INHERITANCE_GRAPH_FULL}` command",
                )
        elif command == CommandsEnum.GENERATE_LINEARIZED_INHERITANCE_GRAPH:
            if params.arguments is not None and len(params.arguments) == 2:
                document_uri = DocumentUri(params.arguments[0])
                canonical_name = str(params.arguments[1])
                context = await self._get_workspace(document_uri)
                return await generate_linearized_inheritance_graph_handler(
                    context, document_uri, canonical_name
                )

        raise LspError(ErrorCodes.InvalidRequest, f"Unknown command: {command}")

    async def get_configuration(self) -> None:
        params = ConfigurationParams(
            items=[
                ConfigurationItem(
                    section="woke",
                )
            ]
        )
        return await self.send_request(
            RequestMethodEnum.WORKSPACE_CONFIGURATION, params
        )



from __future__ import annotations

import enum
import logging
from itertools import chain
from typing import Any, List, Optional, Union

from ...compiler.source_unit_name_resolver import SourceUnitNameResolver
from ..common_structures import (
    Command,
    MarkupContent,
    MarkupKind,
    PartialResultParams,
    Position,
    Range,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    TextEdit,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from ..context import LspContext
from ..lsp_data_model import LspModel
from ..utils import uri_to_path

logger = logging.getLogger(__name__)


class CompletionItemKind(enum.IntEnum):
    TEXT = 1
    METHOD = 2
    FUNCTION = 3
    CONSTRUCTOR = 4
    FIELD = 5
    VARIABLE = 6
    CLASS = 7
    INTERFACE = 8
    MODULE = 9
    PROPERTY = 10
    UNIT = 11
    VALUE = 12
    ENUM = 13
    KEYWORD = 14
    SNIPPET = 15
    COLOR = 16
    FILE = 17
    REFERENCE = 18
    FOLDER = 19
    ENUM_MEMBER = 20
    CONSTANT = 21
    STRUCT = 22
    EVENT = 23
    OPERATOR = 24
    TYPE_PARAMETER = 25


class CompletionItemTag(enum.IntEnum):
    DEPRECATED = 1
    """
    Render a completion as obsolete, usually using a strike-out.
    """


class InsertTextMode(enum.IntEnum):
    AS_IS = 1
    """
    The insertion or replace strings is taken as it is. If the
    value is multi line the lines below the cursor will be
    inserted using the indentation defined in the string value.
    The client will not apply any kind of adjustments to the
    string.
    """
    ADJUST_INDENTATION = 2
    """
    The editor adjusts leading whitespace of new lines so that
    they match the indentation up to the cursor of the line for
    which the item is accepted.
    """


class InsertTextFormat(enum.IntEnum):
    PLAIN_TEXT = 1
    """
    The primary text to be inserted is treated as a plain string.
    """
    SNIPPET = 2
    """
    The primary text to be inserted is treated as a snippet.
    A snippet can define tab stops and placeholders with `$1`, `$2`
    and `${3:foo}`. `$0` defines the final tab stop, it defaults to
    the end of the snippet. Placeholders with equal identifiers are linked,
    that is typing in one will update others too.
    """


#################################################################
# ########## ClientCapabilitiesCompletionItem subclasses ########


class ClientCapabilitiesCompletionItemTagSupport(LspModel):
    value_set: List[CompletionItemTag]


class ClientCapabilitiesCompletionItemResolveSupport(LspModel):
    properties: List[str]
    """
    The properties that a client can resolve lazily.
    """


class ClientCapabilitiesCompletionItemKind(LspModel):
    value_set: Optional[List[CompletionItemKind]]
    """
    If this property is not present the client only supports
    the completion items kinds from `Text` to `Reference` as defined in
    the initial version of the protocol.
    """


class ClientCapabilitiesCompletionItemInsertTextModeSupport(LspModel):
    value_set: List[InsertTextMode]


#################################################################
# ########## CompletionClientCapabilities subclasses ############


class ClientCapabilitiesCompletionItem(LspModel):
    snippet_support: Optional[bool]
    """
    Client supports snippets as insert text.
    A snippet can define tab stops and placeholders with `$1`, `$2`
    and `${3:foo}`. `$0` defines the final tab stop, it defaults to
    the end of the snippet. Placeholders with equal identifiers are
    linked, that is typing in one will update others too.
    """
    commit_characters_support: Optional[bool]
    """
    Client supports commit characters on a completion item.
    """
    documentation_format: Optional[List[MarkupKind]]
    """
    Client supports the following content formats for the documentation
    property. The order describes the preferred format of the client.
    """
    deprecated_support: Optional[bool]
    """
    Client supports the deprecated property on a completion item.
    """
    preselect_support: Optional[bool]
    """
    Client supports the preselect property on a completion item.
    """
    tag_support: Optional[ClientCapabilitiesCompletionItemTagSupport]
    """
    Client supports the tag property on a completion item. Clients
    supporting tags have to handle unknown tags gracefully. Clients
    especially need to preserve unknown tags when sending a completion
    item back to the server in a resolve call.
    """
    insert_replace_support: Optional[bool]
    """
    Client supports insert replace edit to control different behavior if
    a completion item is inserted in the text or should replace text.
    """
    resolve_support: Optional[ClientCapabilitiesCompletionItemResolveSupport]
    """
    Indicates which properties a client can resolve lazily on a
    completion item. Before version 3.16.0 only the predefined properties
    `documentation` and `detail` could be resolved lazily.
    """
    insert_text_mode_support: Optional[
        ClientCapabilitiesCompletionItemInsertTextModeSupport
    ]
    """
    The client supports the `insertTextMode` property on
    a completion item to override the whitespace handling mode
    as defined by the client (see `insertTextMode`).
    """
    label_details_support: Optional[bool]


class ClientCapabilitiesCompletionList(LspModel):
    item_defaults: Optional[List[str]]
    """
    The value lists the supported property names of the
    `CompletionList.itemDefaults` object. If omitted
    no properties are supported.
    """


#################################################################


class CompletionClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    """
    Whether completion supports dynamic registration.
    """
    completion_item: Optional[ClientCapabilitiesCompletionItem]
    """
    The client supports the following `CompletionItem` specific
    capabilities.
    """
    completion_item_kind: Optional[ClientCapabilitiesCompletionItemKind]
    """
    The completion item kind values the client supports. When this
    property exists the client also guarantees that it will
    handle values outside its set gracefully and falls back
    to a default value when unknown.
    """
    context_support: Optional[bool]
    """
    The client supports to send additional context information for a
    `textDocument/completion` request.
    """
    insert_text_mode: Optional[InsertTextMode]
    """
    The client's default when the completion item doesn't provide a
    `insertTextMode` property.
    """
    completion_list: Optional[ClientCapabilitiesCompletionList]
    """
    The client supports the following `CompletionList` specific
    capabilities.
    """


#################################################################
# ########## CompletionOptions subclass #########################


class OptionsCompletionItem(LspModel):
    label_details_support: Optional[bool]
    """
    The server has support for completion item label
    details (see also `CompletionItemLabelDetails`) when receiving
    a completion item in a resolve call.
    """


#################################################################


class CompletionOptions(WorkDoneProgressOptions):
    trigger_characters: Optional[List[str]]
    """
    If code complete should automatically be trigger on characters not being
    valid inside an identifier (for example `.` in JavaScript) list them in
    `triggerCharacters`.
    """
    all_commit_characters: Optional[List[str]]
    """
    The list of all possible characters that commit a completion. This field
    can be used if clients don't support individual commit characters per
    completion item. See client capability
    `completion.completionItem.commitCharactersSupport`.
    """
    resolve_provider: Optional[bool]
    """
    The server provides support to resolve additional
    information for a completion item.
    """
    completion_item: Optional[OptionsCompletionItem]


class CompletionRegistrationOptions(TextDocumentRegistrationOptions, CompletionOptions):
    pass


class CompletionTriggerKind(enum.IntEnum):
    """
    * How a completion was triggered
    """

    INVOKED = 1
    """
    Completion was triggered by typing an identifier
    """
    TRIGGER_CHARACTER = 2
    """
    Completion was triggered by a trigger character specified by
    the `triggerCharacters` properties of the
    `CompletionRegistrationOptions
    """
    TRIGGER_FOR_INCOMPLETE_COMPLETIONS = 3
    """
    Completion was re-triggered as the current completion list is incomplete.
    """


class CompletionContext(LspModel):
    """
    * Contains additional information about the context in which a completion
    request is triggered.
    """

    trigger_kind: CompletionTriggerKind
    """
    How the completion was triggered.
    """
    trigger_character: Optional[str]
    """
    The trigger character (a single character) that has trigger code
    complete. Is undefined if
    `triggerKind !== CompletionTriggerKind.TriggerCharacter`
    """


class CompletionParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    context: Optional[CompletionContext]


class CompletionListItemDefaultsEditRange(LspModel):
    insert: Range
    replace: Range


class CompletionListItemDefaults(LspModel):
    commit_characters: Optional[List[str]]
    """
    A default commit character set.
    """
    edit_range: Optional[Union[Range, CompletionListItemDefaultsEditRange]]
    """
    A default edit range.
    """
    insert_text_format: Optional[InsertTextFormat]
    """
    A default insert text format.
    """
    insert_text_mode: Optional[InsertTextMode]
    """
    A default insert text mode.
    """
    data: Optional[Any]
    """
    A default data value.
    """


class CompletionList(LspModel):
    """
    * Represents a collection of [completion items](#CompletionItem) to be
    presented in the editor.
    """

    is_incomplete: bool
    """
    This list is not complete. Further typing should result in recomputing this list.
    Recomputed lists have all their items replaced (not appended) in the incomplete completion sessions.
    """
    item_defaults: Optional[CompletionListItemDefaults] = None
    """
    In many cases the items of an actual completion result share the same
    value for properties like `commitCharacters` or the range of a text
    edit. A completion list can therefore define item defaults which will
    be used if a completion item itself doesn't specify the value.

    If a completion list specifies a default value and a completion item
    also specifies a corresponding value the one from the item is used.

    Servers are only allowed to return default values if the client
    signals support for this via the `completionList.itemDefaults` capability.
    """
    items: List[CompletionItem]
    """
    The completion items.
    """


class InsertReplaceEdit(LspModel):
    """
    * A special text edit to provide an insert and a replace operation.
    """

    new_text: str
    """
    The string to be inserted.
    """
    insert: Range
    """
    The range if the insert is requested.
    """
    replace: Range
    """
    The range if the replace is requested.
    """


class CompletionItemLabelDetails(LspModel):
    """
    * Additional details for a completion item label.
    """

    detail: Optional[str]
    """
    An optional string which is rendered less prominently directly after
    {@link CompletionItem.label label}, without any spacing. Should be
    used for function signatures or type annotations.
    """
    description: Optional[str]
    """
    An optional string which is rendered less prominently after
    {@link CompletionItemLabelDetails.detail}. Should be used for fully qualified
    names or file path.
    """


class CompletionItem(LspModel):
    label: str
    """
    The label of this completion item.
    The label property is also by default the text that
    is inserted when selecting this completion.
    """
    label_details: Optional[CompletionItemLabelDetails] = None
    """
    Additional details for the label
    """
    kind: Optional[CompletionItemKind] = None
    """
    The kind of this completion item. Based of the kind
    an icon is chosen by the editor. The standardized set
    of available values is defined in `CompletionItemKind`.
    """
    tags: Optional[List[CompletionItemTag]] = None
    """
    Tags for this completion item.
    """
    detail: Optional[str] = None
    """
    A human-readable string with additional information
    about this item, like type or symbol information.
    """
    documentation: Optional[Union[str, MarkupContent]] = None
    """
    A human-readable string that represents a doc-comment.
    """
    deprecated: Optional[bool] = None
    """
    Indicates if this item is deprecated.
    """
    preselect: Optional[bool] = None
    """
    Select this item when showing.
    """
    sort_text: Optional[str] = None
    """
    A string that should be used when comparing this item
    with other items. When `falsy` the label is used
    as the sort text for this item.
    """
    filter_text: Optional[str] = None
    """
    A string that should be used when filtering a set of
    completion items. When `falsy` the label is used as the
    filter text for this item.
    """
    insert_text: Optional[str] = None
    """
    A string that should be inserted into a document when selecting
    this completion. When `falsy` the label is used as the insert text
    for this item.
    """
    insert_text_format: Optional[InsertTextFormat] = None
    """
    The format of the insert text. The format applies to both the
    `insertText` property and the `newText` property of a provided
    `textEdit`. If omitted defaults to `InsertTextFormat.PlainText`.
    """
    insert_text_mode: Optional[InsertTextMode] = None
    """
    How whitespace and indentation is handled during completion
    item insertion. If not provided the client's default value depends on
    the `textDocument.completion.insertTextMode` client capability.
    """
    text_edit: Optional[Union[TextEdit, InsertReplaceEdit]] = None
    """
    An edit which is applied to a document when selecting this completion.
    When an edit is provided the value of `insertText` is ignored.
    """
    text_edit_text: Optional[str] = None
    """
    The edit text ussed if the completion item is part of a
    CompletionList and CompletionList defines an item default for
    the text edit range.
    """
    additional_text_edits: Optional[List[TextEdit]] = None
    """
    An optional array of additional text edits that are applied when
    selecting this completion. Edits must not overlap (including the same
    insert position) with the main edit nor with themselves.
    """
    commit_characters: Optional[List[str]] = None
    """
    An optional set of characters that when pressed while this completion is
    active will accept it first and then type that character. *Note* that all
    commit characters should have `length=1` and that superfluous characters
    will be ignored.
    """
    command: Optional[Command] = None
    """
    An optional command that is executed *after* inserting this completion.
    *Note* that additional modifications to the current document should be
    described with the additionalTextEdits-property.
    """
    # LSPAny
    data: Optional[Any] = None
    """
    A data entry field that is preserved on a completion item between
    a completion and a completion resolve request.
    """


async def completion(
    context: LspContext, params: CompletionParams
) -> Optional[CompletionList]:
    path = uri_to_path(params.text_document.uri).resolve()
    try:
        node = await context.parser.get_node_at_position(
            path, params.position.line, params.position.character
        )
    except KeyError:
        return None

    while node is not None:
        if node.type == "import_directive":
            break
        node = node.parent

    if node is None:
        return None

    this_source_unit_name = None
    for include_path in chain(
        context.config.compiler.solc.include_paths, [context.config.project_root_path]
    ):
        try:
            rel_path = str(path.relative_to(include_path).as_posix())
            if this_source_unit_name is None or len(this_source_unit_name) > len(
                rel_path
            ):
                this_source_unit_name = rel_path
        except ValueError:
            continue

    if this_source_unit_name is None:
        return None

    source_node = node.child_by_field_name("source")
    import_str = source_node.text.decode("utf-16-le")[1:-1]  # remove quotes
    resolver = SourceUnitNameResolver(context.config)
    import_str = resolver.apply_remapping(this_source_unit_name, import_str)

    parts = import_str.split("/")
    parent = "/".join(parts[:-1])
    prefix = parts[-1]

    completions = set()

    if len(parent) == 0:
        for p in path.parent.iterdir():
            if p.is_dir():
                completions.add(("./" + p.name + "/", CompletionItemKind.FOLDER))
            elif p.is_file() and p.suffix == ".sol" and p != path:
                completions.add(("./" + p.name, CompletionItemKind.FILE))

        for include_path in chain(
            context.config.compiler.solc.include_paths,
            [context.config.project_root_path],
        ):
            if include_path.is_dir():
                for p in include_path.iterdir():
                    if p.is_dir():
                        completions.add((p.name + "/", CompletionItemKind.FOLDER))
                    elif p.is_file() and p.suffix == ".sol":
                        completions.add((p.name, CompletionItemKind.FILE))

        for remapping in context.config.compiler.solc.remappings:
            if remapping.context is None or this_source_unit_name.startswith(
                remapping.context
            ):
                completions.add((remapping.prefix, CompletionItemKind.MODULE))

    elif parent.startswith("."):
        dir = path.parent / parent
        if dir.is_dir():
            for p in dir.iterdir():
                if p.is_dir():
                    completions.add((p.name + "/", CompletionItemKind.FOLDER))
                elif p.is_file() and p.suffix == ".sol" and p != path:
                    completions.add((p.name, CompletionItemKind.FILE))
    else:
        for include_path in chain(
            context.config.compiler.solc.include_paths,
            [context.config.project_root_path],
        ):
            if include_path.is_dir():
                dir = include_path / parent
                if dir.is_dir():
                    for p in dir.iterdir():
                        if p.is_dir():
                            completions.add((p.name + "/", CompletionItemKind.FOLDER))
                        elif p.is_file() and p.suffix == ".sol":
                            completions.add((p.name, CompletionItemKind.FILE))

    import_end_pos = Position(
        line=source_node.end_point[0],
        character=source_node.end_point[1] // 2 - 1,
    )

    return CompletionList(
        is_incomplete=False,
        items=[
            CompletionItem(
                label=l,
                kind=k,
                text_edit=TextEdit(
                    range=Range(
                        start=import_end_pos,
                        end=import_end_pos,
                    ),
                    new_text=l[len(prefix) :],
                ),
            )
            for l, k in sorted(completions)
            if l.startswith(prefix)
        ],
    )


CompletionList.update_forward_refs()



from __future__ import annotations

import logging
import re
from collections import deque
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Deque, Iterator, List, Optional, Set, Tuple

from ..expressions.identifier import Identifier
from ..reference_resolver import CallbackParams

if TYPE_CHECKING:
    from .source_unit import SourceUnit

from woke.ir.ast import AstNodeId, SolcImportDirective

from ..abc import IrAbc, SolidityAbc
from ..utils import IrInitTuple

logger = logging.getLogger(__name__)

ENCODING = "utf-8"

FILENAME = r"""(?P<filename>'.*[^\\]'|".*[^\\]")"""
SYMBOL = r"[_a-zA-Z][_a-zA-Z0-9]*"
ALIAS = r"\s*{symbol}(?:\s+as\s+{symbol})?\s*".format(symbol=SYMBOL)
IMPORT_FILENAME_RE = re.compile(
    r"import\s*{filename}\s*".format(filename=FILENAME).encode(ENCODING)
)
IMPORT_AS_FROM_RE = re.compile(
    r"import\s*\*\s*as\s+{symbol}\s+from\s*{filename}\s*".format(
        filename=FILENAME, symbol=SYMBOL
    ).encode(ENCODING)
)
IMPORT_AS_RE = re.compile(
    r"import\s*{filename}\s*as\s+{symbol}\s*".format(
        filename=FILENAME, symbol=SYMBOL
    ).encode(ENCODING)
)
IMPORT_ALIAS_LIST = re.compile(
    r"import\s*{{{alias}(?:,{alias})*}}\s*from\s*{filename}\s*".format(
        alias=ALIAS, filename=FILENAME
    ).encode(ENCODING)
)


@dataclass
class SymbolAlias:
    """
    Helper class representing a symbol alias in an import directive of the`:::solidity import {symbol as alias} from "file.sol";` form.

    !!! example
        `symbol` is the `foreign` attribute and `alias` is the `local` attribute in the following example:
        ```solidity
        import {symbol as alias} from "file.sol";
        ```

    Attributes:
        foreign (Identifier): Identifier referencing the symbol in the imported file.
        local (Optional[str]): Alias name of the imported symbol (if any).
    """

    foreign: Identifier
    local: Optional[str]


class ImportDirective(SolidityAbc):
    """
    !!! example
        ```solidity
        import "SafeLib.sol";
        ```
        ```solidity
        import "SafeLib.sol" as SafeLib;
        ```
        ```solidity
        import * as SafeLib from "SafeLib.sol";
        ```
        ```solidity
        import { SafeType as CustomSafeType } from "SafeLib.sol";
        ```
    """

    _ast_node: SolcImportDirective
    _parent: SourceUnit

    _source_unit_name: str
    _import_string: str
    _imported_file: Path
    _source_unit_id: AstNodeId
    _symbol_aliases: List[SymbolAlias]
    _unit_alias: Optional[str]

    def __init__(
        self,
        init: IrInitTuple,
        import_directive: SolcImportDirective,
        parent: SolidityAbc,
    ):
        super().__init__(init, import_directive, parent)
        self._source_unit_name = import_directive.absolute_path
        self._import_string = import_directive.file
        self._imported_file = init.cu.source_unit_name_to_path(self._source_unit_name)
        self._source_unit_id = import_directive.source_unit
        self._symbol_aliases = []
        if len(import_directive.unit_alias) > 0:
            self._unit_alias = import_directive.unit_alias
        else:
            self._unit_alias = None

        for alias in import_directive.symbol_aliases:
            self._symbol_aliases.append(
                SymbolAlias(Identifier(init, alias.foreign, self), alias.local)
            )
        self._reference_resolver.register_post_process_callback(
            self._post_process, priority=-2
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for symbol_alias in self._symbol_aliases:
            yield from symbol_alias.foreign

    def _post_process(self, callback_params: CallbackParams):
        # referenced declaration ID is missing (for whatever reason) in import directive symbol aliases
        # for example `import { SafeType } from "SafeLib.sol";`
        # fix: find these reference IDs manually
        # seems to be fixed in solc >= 0.8.12
        for symbol_alias in self._symbol_aliases:
            if symbol_alias.foreign._referenced_declaration_id is not None:
                continue

            source_units_queue: Deque[SourceUnit] = deque(
                [callback_params.source_units[self._imported_file]]
            )
            processed_source_units: Set[Path] = {self._imported_file}
            referenced_declaration = None
            searched_name = symbol_alias.foreign.name

            while source_units_queue and referenced_declaration is None:
                imported_source_unit = source_units_queue.pop()

                for declaration in imported_source_unit.declarations_iter():
                    if declaration.canonical_name == searched_name:
                        referenced_declaration = declaration
                        break

                for import_ in imported_source_unit.imports:
                    if import_.unit_alias == searched_name:
                        referenced_declaration = import_
                        break

                    # handle the case when an imported symbol is an alias of another symbol
                    for alias in import_.symbol_aliases:
                        if alias.local == symbol_alias.foreign.name:
                            searched_name = alias.foreign.name
                    if import_.imported_file not in processed_source_units:
                        source_units_queue.append(
                            callback_params.source_units[import_.imported_file]
                        )
                        processed_source_units.add(import_.imported_file)

            assert referenced_declaration is not None
            node_path_order = self._reference_resolver.get_node_path_order(
                AstNodeId(referenced_declaration.ast_node_id),
                referenced_declaration.cu_hash,
            )
            referenced_declaration_id = (
                self._reference_resolver.get_ast_id_from_cu_node_path_order(
                    node_path_order, self.cu_hash
                )
            )
            symbol_alias.foreign._referenced_declaration_id = referenced_declaration_id

    @property
    def parent(self) -> SourceUnit:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def source_unit_name(self) -> str:
        """
        Returns:
            Source unit name of the imported file.
        """
        return self._source_unit_name

    @property
    def imported_file(self) -> Path:
        """
        Returns:
            Absolute path of the imported file.
        """
        return self._imported_file

    @property
    def import_string(self) -> str:
        """
        Returns:
            Import string as it appears in the source code.
        """
        return self._import_string

    @property
    def source_unit(self) -> SourceUnit:
        """
        Returns:
            Source unit imported by this import directive.
        """
        from .source_unit import SourceUnit

        node = self._reference_resolver.resolve_node(
            self._source_unit_id, self._cu_hash
        )
        assert isinstance(node, SourceUnit)
        return node

    @property
    def symbol_aliases(self) -> Tuple[SymbolAlias, ...]:
        """
        Is only set in the case of `:::solidity import { SafeType as CustomSafeType } from "SafeLib.sol";` import directive type.
        Returns:
            Symbol aliases present in the import directive.
        """
        return tuple(self._symbol_aliases)

    @property
    def unit_alias(self) -> Optional[str]:
        """
        !!! example
            Is `SafeLib` in the case of these import directives:
            ```solidity
            import "SafeLib.sol" as SafeLib;
            ```
            ```solidity
            import * as SafeLib from "SafeLib.sol";
            ```

            Is `None` in the case of these import directives:
            ```solidity
            import "SafeLib.sol";
            ```
            ```solidity
            import { SafeType as CustomSafeType } from "SafeLib.sol";
            ```
        Returns:
            Alias for the namespace of the imported source unit.
        """
        return self._unit_alias

    @property
    @lru_cache(maxsize=2048)
    def import_string_pos(self) -> Tuple[int, int]:
        """
        Returns:
            Byte offsets (start, end) of the import string in the source file.
        """
        source_start = self._ast_node.src.byte_offset

        res = (
            IMPORT_FILENAME_RE,
            IMPORT_AS_FROM_RE,
            IMPORT_AS_RE,
            IMPORT_ALIAS_LIST,
        )
        matches = list(re.match(self._source) for re in res)
        assert any(matches)
        match = next(m for m in matches if m)
        return source_start + match.start("filename"), source_start + match.end(
            "filename"
        )



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple

from woke.ir.enums import ModifiesStateFlag

from ..statements.block import Block
from ..utils import IrInitTuple
from .parameter_list import ParameterList

if TYPE_CHECKING:
    from ..statements.try_statement import TryStatement

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcTryCatchClause


class TryCatchClause(SolidityAbc):
    """
    !!! example
        - `:::solidity returns(uint x) {}`,
        - `:::solidity catch Error(string memory reason) {}`,
        - `:::solidity catch Panic(uint errorCode) {}`,
        - `:::solidity catch (bytes memory lowLevelData) {}` are all try/catch clauses in the following example:
        ```solidity
        contract C {
            function foo() public view {
                try this.bar(10) returns(uint x) {}
                catch Error(string memory reason) {}
                catch Panic(uint errorCode) {}
                catch (bytes memory lowLevelData) {}
            }

            function bar(uint x) external pure returns(uint) {
                return x;
            }
        }
        ```
    """

    _ast_node: SolcTryCatchClause
    _parent: TryStatement

    _block: Block
    _error_name: str
    _parameters: Optional[ParameterList]

    def __init__(
        self,
        init: IrInitTuple,
        try_catch_clause: SolcTryCatchClause,
        parent: TryStatement,
    ):
        super().__init__(init, try_catch_clause, parent)
        self._block = Block(init, try_catch_clause.block, self)
        self._error_name = try_catch_clause.error_name

        if try_catch_clause.parameters is None:
            self._parameters = None
        else:
            self._parameters = ParameterList(init, try_catch_clause.parameters, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._block
        if self._parameters is not None:
            yield from self._parameters

    @property
    def parent(self) -> TryStatement:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def block(self) -> Block:
        """
        Returns:
            Body of the try/catch clause.
        """
        return self._block

    @property
    def error_name(self) -> str:
        """
        !!! example
            For the following snippet:
            ```solidity
            try this.f() returns (uint256) {
                // ...
            } catch Error(string memory reason) {
                // ...
            } catch Panic(uint errorCode) {
                // ...
            } catch (bytes memory lowLevelData) {
                // ...
            }
            ```

            - the `error_name` of the first (try) clause is empty,
            - the `error_name` of the second (catch) clause is `Error`,
            - the `error_name` of the third (catch) clause is `Panic`,
            - the `error_name` of the fourth (catch) clause is empty.
        Returns:
            Error name of the try/catch clause.
        """
        return self._error_name

    @property
    def parameters(self) -> Optional[ParameterList]:
        """
        Can be `None` if the try clause does not have return parameters or if the catch clause does not accept parameters.
        !!! example
            Both clauses in the following example do not have parameters:
            ```solidity
            try this.f() {
                // ...
            } catch {
                // ...
            }
            ```

        !!! example
            `:::solidity (uint x)`, `:::solidity (string memory reason)`, `:::solidity (uint errorCode)` and `:::solidity (bytes memory lowLevelData)` are the parameters of the try/catch clauses in the following example:
            ```solidity
            try this.f() returns (uint x) {
                // ...
            } catch Error(string memory reason) {
                // ...
            } catch Panic(uint errorCode) {
                // ...
            } catch (bytes memory lowLevelData) {
                // ...
            }
            ```

        Returns:
            Return parameters in the case of a try clause, or error parameters in the case of a catch clause.
        """
        return self._parameters

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return self.block.modifies_state



from __future__ import annotations

from abc import ABC
from typing import Iterator, Set, Tuple

from woke.ir.ast import YulNode
from woke.ir.enums import ModifiesStateFlag

from ..abc import IrAbc


class YulAbc(IrAbc, ABC):
    """
    Abstract base class for all Yul IR nodes.
    """

    _ast_node: YulNode

    def __iter__(self) -> Iterator[YulAbc]:
        yield self

    @property
    def ast_node(self) -> YulNode:
        return self._ast_node

    @property
    # @abstractmethod
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return set()  # TODO


class YulStatementAbc(YulAbc, ABC):
    """
    Abstract base class for all Yul IR statements.
    """

    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Union

from woke.ir.ast import (
    SolcYulExpressionStatement,
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulLiteral,
)
from woke.ir.utils import IrInitTuple

from .abc import YulAbc, YulStatementAbc
from .function_call import YulFunctionCall
from .identifier import YulIdentifier
from .literal import YulLiteral

if TYPE_CHECKING:
    from .block import YulBlock


class YulExpressionStatement(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _expression: Union[YulFunctionCall, YulIdentifier, YulLiteral]

    def __init__(
        self,
        init: IrInitTuple,
        expression_statement: SolcYulExpressionStatement,
        parent: YulAbc,
    ):
        super().__init__(init, expression_statement, parent)
        if isinstance(expression_statement.expression, SolcYulFunctionCall):
            self._expression = YulFunctionCall(
                init, expression_statement.expression, self
            )
        elif isinstance(expression_statement.expression, SolcYulIdentifier):
            self._expression = YulIdentifier(
                init, expression_statement.expression, self
            )
        elif isinstance(expression_statement.expression, SolcYulLiteral):
            self._expression = YulLiteral(init, expression_statement.expression, self)
        else:
            assert False, f"Unexpected type: {type(expression_statement.expression)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._expression

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def expression(self) -> Union[YulFunctionCall, YulIdentifier, YulLiteral]:
        return self._expression



from pathlib import Path
from typing import List, Optional

import rich_click as click
from click.core import Context
from rich_click.rich_group import RichGroup

from .console import console


class NewCommandAlias(RichGroup):
    def get_command(self, ctx: Context, cmd_name: str):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        if cmd_name in {"new", "add", "create"}:
            return self.get_command(ctx, "new")
        if cmd_name in {"remove", "delete", "rm"}:
            return self.get_command(ctx, "remove")
        return None

    def resolve_command(self, ctx: Context, args: List[str]):
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name if cmd is not None else None, cmd, args


@click.group(name="accounts", cls=NewCommandAlias)
@click.pass_context
def run_accounts(ctx: Context):
    """
    Run Woke accounts manager.
    """
    from eth_account import Account

    from woke.config import WokeConfig

    config = WokeConfig()
    config.load_configs()
    ctx.obj["config"] = config

    Account.enable_unaudited_hdwallet_features()


@run_accounts.command(name="new")
@click.option("--kdf", default="scrypt", help="Key derivation function to encrypt key")
@click.option(
    "--password",
    prompt="Password to encrypt key",
    confirmation_prompt=True,
    default="",
    hide_input=True,
    help="Password to encrypt key",
)
@click.option(
    "--mnemonic/--no-mnemonic", is_flag=True, default=True, help="Print mnemonic"
)
@click.option("--language", "--lang", default="english", help="Mnemonic language")
@click.option("--words", default=12, help="Number of words")
@click.option("--hd-path", default="m/44'/60'/0'/0/0", help="HD path")
@click.option("--keystore", type=click.Path(file_okay=False), help="Keystore path")
@click.argument("alias")
@click.pass_context
def accounts_new(
    ctx: Context,
    kdf: str,
    password: str,
    mnemonic: bool,
    language: str,
    words: int,
    hd_path: str,
    keystore: Optional[str],
    alias: str,
):
    """
    Create a new account.
    """
    import json

    from eth_account import Account
    from eth_utils.address import to_checksum_address

    if keystore is None:
        path = Path(ctx.obj["config"].global_data_path) / "keystore"
    else:
        path = Path(keystore)
    if not path.exists():
        path.mkdir(parents=True)

    if not path.is_dir():
        raise click.BadParameter("Keystore path must be a directory")

    path = path / f"{alias}.json"
    if path.exists():
        raise click.BadParameter("Alias already exists, remove it first")

    if mnemonic:
        passphrase = click.prompt(
            "Mnemonic passphrase", default="", hide_input=True, confirmation_prompt=True
        )
        acc, mnemonic_str = Account.create_with_mnemonic(
            passphrase, words, language, hd_path
        )
        if click.confirm("Mnemonic will be printed, continue?"):
            console.print(f"[bold]{mnemonic_str}[/bold]")
    else:
        acc = Account.create()

    k = Account.encrypt(acc.key, password, kdf)

    with path.open("w") as f:
        json.dump(k, f)

    console.print(
        f"[green]Account created: {alias} {to_checksum_address(acc.address)}[/green]"
    )


@run_accounts.command(name="remove")
@click.option(
    "--keystore", type=click.Path(exists=True, file_okay=False), help="Keystore path"
)
@click.argument("alias")
@click.pass_context
def accounts_remove(ctx: Context, keystore: Optional[str], alias: str):
    """
    Remove an account.
    """
    import json

    from eth_utils.address import to_checksum_address

    if keystore is None:
        path = Path(ctx.obj["config"].global_data_path) / "keystore"
    else:
        path = Path(keystore)
    if not path.is_dir():
        raise click.BadParameter("Keystore path must be a directory")

    path = path / f"{alias}.json"
    if not path.exists():
        raise click.BadParameter("Alias does not exist")

    data = json.loads(path.read_text())
    path.unlink()
    console.print(
        f"[green]Account {to_checksum_address(data['address'])} removed[/green]"
    )


@run_accounts.command(name="list")
@click.option(
    "--keystore", type=click.Path(exists=True, file_okay=False), help="Keystore path"
)
@click.pass_context
def accounts_list(ctx: Context, keystore: Optional[str]):
    """
    List all accounts.
    """
    import json

    from eth_utils.address import to_checksum_address

    if keystore is None:
        path = Path(ctx.obj["config"].global_data_path) / "keystore"
    else:
        path = Path(keystore)
    if not path.is_dir():
        return

    for file in sorted(path.iterdir()):
        if file.suffix == ".json" and file.is_file():
            with file.open() as f:
                k = json.load(f)
            console.print(f"{file.stem}: {to_checksum_address(k['address'])}")


@run_accounts.command(name="import")
@click.option("--kdf", default="scrypt", help="Key derivation function to encrypt key")
@click.option(
    "--password",
    prompt="Password to encrypt key",
    confirmation_prompt=True,
    default="",
    hide_input=True,
    help="Password to encrypt key",
)
@click.option(
    "--hd-path", default="m/44'/60'/0'/0/0", help="HD path when importing mnemonic"
)
@click.option("--keystore", type=click.Path(file_okay=False), help="Keystore path")
@click.argument("alias")
@click.pass_context
def accounts_import(
    ctx: Context,
    kdf: str,
    password: str,
    hd_path: str,
    keystore: Optional[str],
    alias: str,
):
    """
    Import an account from a private key or mnemonic.
    """
    import json

    from eth_account import Account
    from eth_utils.address import to_checksum_address

    if keystore is None:
        path = Path(ctx.obj["config"].global_data_path) / "keystore"
    else:
        path = Path(keystore)
    if not path.exists():
        path.mkdir(parents=True)

    if not path.is_dir():
        raise click.BadParameter("Keystore path must be a directory")

    path = path / f"{alias}.json"
    if path.exists():
        raise click.BadParameter("Alias already exists, remove it first")

    key_or_mnemonic = click.prompt("Private key or mnemonic", hide_input=True)
    key = None

    if key_or_mnemonic.startswith("0x"):
        try:
            key = bytes.fromhex(key_or_mnemonic[2:])
        except ValueError:
            raise click.BadParameter("Invalid private key")

    if key is None:
        try:
            key = bytes.fromhex(key_or_mnemonic)
        except ValueError:
            key = None

    if key is not None:
        acc = Account.from_key(key)
    else:
        acc = Account.from_mnemonic(key_or_mnemonic, "", hd_path)

    k = Account.encrypt(acc.key, password, kdf)

    with path.open("w") as f:
        json.dump(k, f)

    console.print(
        f"[green]Account imported: {alias} {to_checksum_address(acc.address)}[/green]"
    )


@run_accounts.command(name="export")
@click.option(
    "--keystore", type=click.Path(exists=True, file_okay=False), help="Keystore path"
)
@click.argument("alias")
@click.pass_context
def accounts_export(ctx: Context, keystore: Optional[str], alias: str):
    """
    Export an account's private key.
    """
    import json

    from eth_account import Account
    from eth_utils.address import to_checksum_address

    if keystore is None:
        path = Path(ctx.obj["config"].global_data_path) / "keystore"
    else:
        path = Path(keystore)
    if not path.is_dir():
        raise click.BadParameter("Keystore path must be a directory")

    path = path / f"{alias}.json"
    if not path.exists():
        raise click.BadParameter("Alias does not exist")

    with path.open() as f:
        k = json.load(f)

    x = Account.decrypt(k, click.prompt("Password", default="", hide_input=True))
    console.print(
        f"[green]Address {to_checksum_address(k['address'])} with private key {x.hex()}[/green]"
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from woke.ir.ast import SolcYulTypedName
from woke.ir.utils import IrInitTuple

from .abc import YulAbc

if TYPE_CHECKING:
    from .function_definition import YulFunctionDefinition
    from .variable_declaration import YulVariableDeclaration


class YulTypedName(YulAbc):
    """
    TBD
    """

    _parent: Union[YulFunctionDefinition, YulVariableDeclaration]
    _name: str
    _type: str

    def __init__(self, init: IrInitTuple, typed_name: SolcYulTypedName, parent: YulAbc):
        super().__init__(init, typed_name, parent)
        self._name = typed_name.name
        self._type = typed_name.type

    @property
    def parent(self) -> Union[YulFunctionDefinition, YulVariableDeclaration]:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Union

from .elementary_type_name import ElementaryTypeName
from .user_defined_type_name import UserDefinedTypeName

if TYPE_CHECKING:
    from ..declarations.variable_declaration import VariableDeclaration
    from ..meta.using_for_directive import UsingForDirective
    from .array_type_name import ArrayTypeName

import woke.ir.types as types
from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcMapping
from woke.ir.type_names.abc import TypeNameAbc
from woke.ir.utils import IrInitTuple


class Mapping(TypeNameAbc):
    """
    Mapping type name.

    !!! example
        A mapping type name can be used:

        - inside a [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration]:
            - `:::solidity mapping(address => uint)` in line 1,
            - `:::solidity mapping(address => mapping(address => uint))` in line 8,
        - inside a [UsingForDirective][woke.ir.meta.using_for_directive.UsingForDirective]:
            - `:::solidity mapping(address => uint)` in line 5,
        - inside an [ArrayTypeName][woke.ir.type_names.array_type_name.ArrayTypeName]:
            - `:::solidity mapping(address => uint)` in line 9,
        - inside a [Mapping][woke.ir.type_names.mapping.Mapping]:
            - `:::solidity mapping(address => uint)` in line 8.

        ```solidity linenums="1"
        function remove(mapping(address => uint) storage balances, address account) {
            delete balances[account];
        }

        using {remove} for mapping(address => uint);

        contract C {
            mapping(address => mapping(address => uint)) public allowances;
            mapping(address => uint)[2] public balances;
        }
        ```
    """

    _ast_node: SolcMapping
    _parent: Union[VariableDeclaration, UsingForDirective, ArrayTypeName, Mapping]

    _key_type: Union[ElementaryTypeName, UserDefinedTypeName]
    _value_type: TypeNameAbc

    def __init__(self, init: IrInitTuple, mapping: SolcMapping, parent: SolidityAbc):
        super().__init__(init, mapping, parent)
        key_type = TypeNameAbc.from_ast(init, mapping.key_type, self)
        assert isinstance(key_type, (ElementaryTypeName, UserDefinedTypeName))
        self._key_type = key_type
        self._value_type = TypeNameAbc.from_ast(init, mapping.value_type, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._key_type
        yield from self._value_type

    @property
    def parent(
        self,
    ) -> Union[VariableDeclaration, UsingForDirective, ArrayTypeName, Mapping]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def type(self) -> types.Mapping:
        """
        Returns:
            Type description.
        """
        t = super().type
        assert isinstance(t, types.Mapping)
        return t

    @property
    def key_type(self) -> Union[ElementaryTypeName, UserDefinedTypeName]:
        """
        Can only be:

        - an [ElementaryTypeName][woke.ir.type_names.elementary_type_name.ElementaryTypeName],
        - a [UserDefinedTypeName][woke.ir.type_names.user_defined_type_name.UserDefinedTypeName] of a [Contract][woke.ir.types.Contract] type,
        - a [UserDefinedTypeName][woke.ir.type_names.user_defined_type_name.UserDefinedTypeName] of an [Enum][woke.ir.types.Enum] type,
        - a [UserDefinedTypeName][woke.ir.type_names.user_defined_type_name.UserDefinedTypeName] of a [UserDefinedValueType][woke.ir.types.UserDefinedValueType] type.
        Returns:
            Mapping key type name.
        """
        return self._key_type

    @property
    def value_type(self) -> TypeNameAbc:
        """
        Returns:
            Mapping value type name.
        """
        return self._value_type



TEMPLATE = """from __future__ import annotations

import rich_click as click
import woke.ir as ir
import woke.ir.types as types
from woke.printers import Printer, printer


class {class_name}(Printer):
    def print(self) -> None:
        pass

    @printer.command(name="{command_name}")
    def cli(self) -> None:
        pass
"""



from __future__ import annotations

from typing import List, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


def check_missing_return(node: ir.FunctionDefinition) -> List[DetectorResult]:
    from woke.analysis.utils import check_all_return_params_set

    cfg = node.cfg
    assert cfg is not None
    end = cfg.end_block
    graph = cfg.graph
    detections = []

    for block in graph.predecessors(end):
        has_return = False
        for stmt in block.statements:
            if isinstance(stmt, ir.ExpressionStatement):
                if (
                    isinstance(stmt.expression, ir.FunctionCall)
                    and stmt.expression.function_called
                    == ir.enums.GlobalSymbolsEnum.REVERT
                ):
                    has_return = True
                    break
                elif isinstance(stmt.expression, ir.Return) or isinstance(
                    stmt.expression, ir.RevertStatement
                ):
                    has_return = True
                    break
            elif isinstance(stmt, ir.Return) or isinstance(stmt, ir.RevertStatement):
                has_return = True
                break
            elif isinstance(stmt, ir.InlineAssembly) and stmt.yul_block is not None:
                for yul_stmt in stmt.yul_block.statements:
                    if (
                        isinstance(yul_stmt, ir.YulExpressionStatement)
                        and isinstance(yul_stmt.expression, ir.YulFunctionCall)
                        and yul_stmt.expression.function_name.name
                        in ("revert", "return")
                    ):
                        has_return = True
                        break
        if not has_return:
            has_return_params_named = True
            for param in node.return_parameters.parameters:
                if param.name == "":
                    has_return_params_named = False
                    break

            if not has_return_params_named:
                detections.append(
                    DetectorResult(
                        Detection(
                            node,
                            "Not all code paths have return or revert statement",
                            lsp_range=node.name_location,
                        ),
                        DetectionImpact.MEDIUM,
                        DetectionConfidence.LOW,
                    )
                )
            else:
                solved, params = check_all_return_params_set(
                    node.return_parameters.parameters, graph, block, cfg.start_block
                )
                if not solved:
                    detections.append(
                        DetectorResult(
                            Detection(
                                node,
                                "Not all code paths have return or revert statement and the return values "
                                "are not set either",
                                lsp_range=node.name_location,
                            ),
                            DetectionImpact.MEDIUM,
                            DetectionConfidence.LOW,
                        )
                    )
    return detections


class MissingReturnDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_definition(self, node: ir.FunctionDefinition):
        if (
            node.body is None
            or len(node.body.statements) == 0
            or len(node.return_parameters.parameters) == 0
        ):
            return

        for det in check_missing_return(node):
            self._detections.add(det)

    @detector.command("missing-return")
    def cli(self):
        """
        Detector that checks if all possible paths have a return or revert statement or have all
        return values set
        """
        pass



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple

from woke.core.visitor import Visitor
from woke.utils import StrEnum

if TYPE_CHECKING:
    from woke.ir.abc import IrAbc


class DetectionImpact(StrEnum):
    INFO = "info"
    WARNING = "warning"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class DetectionConfidence(StrEnum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@dataclass(eq=True, frozen=True)
class Detection:
    ir_node: IrAbc
    message: str
    subdetections: Tuple[Detection, ...] = field(default_factory=tuple)
    lsp_range: Optional[Tuple[int, int]] = field(default=None)
    subdetections_mandatory: bool = field(default=True)


@dataclass(eq=True, frozen=True)
class DetectorResult:
    detection: Detection
    impact: DetectionImpact
    confidence: DetectionConfidence
    url: Optional[str] = field(default=None)

    def __post_init__(self):
        if self.impact not in DetectionImpact.__members__.values():
            raise ValueError(f"Invalid impact: {self.impact}")
        if self.confidence not in DetectionConfidence.__members__.values():
            raise ValueError(f"Invalid confidence: {self.confidence}")


class Detector(Visitor, metaclass=ABCMeta):
    paths: List[Path]

    @abstractmethod
    def detect(self) -> List[DetectorResult]:
        ...



from contextlib import contextmanager
from os import chdir
from pathlib import Path
from typing import Set, Union


@contextmanager
def change_cwd(path: Union[str, Path]):
    """
    Temporary change the current working directory to the path provided as the first argument.
    """
    orig_cwd = Path.cwd().resolve()
    try:
        chdir(Path(path).resolve())
        yield
    finally:
        chdir(orig_cwd)


@contextmanager
def recursion_guard(guard: Set, *args):
    try:
        guard.add(args)
        yield
    finally:
        guard.remove(args)



from __future__ import annotations

from typing import List, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class UnusedContractDetector(Detector):
    _detections: Set[Detection]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return [
            DetectorResult(
                d, impact=DetectionImpact.WARNING, confidence=DetectionConfidence.HIGH
            )
            for d in self._detections
        ]

    def visit_contract_definition(self, node: ir.ContractDefinition):
        if len(node.references) == 0:
            if node.abstract:
                self._detections.add(
                    Detection(node, "Contract not used", lsp_range=node.name_location)
                )
            elif node.kind == ir.enums.ContractKind.INTERFACE:
                self._detections.add(
                    Detection(node, "Interface not used", lsp_range=node.name_location)
                )
            elif node.kind == ir.enums.ContractKind.LIBRARY:
                self._detections.add(
                    Detection(node, "Library not used", lsp_range=node.name_location)
                )

    @detector.command("unused-contract")
    def cli(self):
        """
        Detects abstract contracts, interfaces and libraries that are not used.
        """
        pass



import dataclasses
import enum
import random
import string
from typing import Any, Callable, Dict, Optional, Type

from typing_extensions import Annotated, get_args, get_origin, get_type_hints

import woke.deployment
import woke.testing
from woke.development.core import Account, Address, Chain, NotConnectedError, Wei
from woke.development.primitive_types import Length, ValueRange


def random_account(
    *,
    lower_bound: int = 0,
    upper_bound: Optional[int] = None,
    predicate: Optional[Callable[[Account], bool]] = None,
    chain: Optional[Chain] = None,
) -> Account:
    if chain is None:
        if (
            woke.deployment.default_chain.connected
            and woke.testing.default_chain.connected
        ):
            raise ValueError(
                "Both default_chain and woke.deployment.default_chain are connected. Please specify which chain to use."
            )
        if woke.deployment.default_chain.connected:
            chain = woke.deployment.default_chain
        elif woke.testing.default_chain.connected:
            chain = woke.testing.default_chain
        else:
            raise NotConnectedError("default_chain not connected")

    accounts = chain.accounts
    if upper_bound is None:
        upper_bound = len(accounts)
    accounts = accounts[lower_bound:upper_bound]
    if predicate is not None:
        accounts = [acc for acc in accounts if predicate(acc)]
    return random.choice(accounts)


def random_address(*, zero_address_prob: float = 0) -> Address:
    if zero_address_prob is not None and random.random() < zero_address_prob:
        return Address(0)
    ret = Address("0x" + random_bytes(20).hex())
    while ret == Address(0):
        ret = Address("0x" + random_bytes(20).hex())
    return ret


def random_int(
    min: int,
    max: int,
    *,
    min_prob: Optional[float] = None,
    zero_prob: Optional[float] = None,
    max_prob: Optional[float] = None,
    edge_values_prob: Optional[float] = None,
) -> int:
    p = random.random()

    if min_prob is None:
        if edge_values_prob is not None:
            min_prob = edge_values_prob

    if min_prob is not None:
        if p < min_prob:
            return min
        p -= min_prob
        min += 1

    if min < 0 < max:
        if zero_prob is None:
            if edge_values_prob is not None:
                zero_prob = edge_values_prob
    else:
        zero_prob = None

    if zero_prob is not None:
        if p < zero_prob:
            return 0
        p -= zero_prob

    if max_prob is None:
        if edge_values_prob is not None:
            max_prob = edge_values_prob

    if max_prob is not None:
        if p < max_prob:
            return max
        max -= 1

    ret = random.randint(min, max)
    while zero_prob is not None and ret == 0:
        ret = random.randint(min, max)
    return ret


def random_bool(
    *,
    true_prob: float = 0.5,
) -> bool:
    return random.random() < true_prob


def random_string(
    min: int,
    max: int,
    *,
    alphabet: str = string.printable,
    predicate: Optional[Callable[[str], bool]] = None,
) -> str:
    def gen() -> str:
        len = random.randint(min, max)
        return "".join(random.choice(alphabet) for _ in range(len))

    ret = gen()
    if predicate is None:
        return ret

    while not predicate(ret):
        ret = gen()
    return ret


def random_bytes(
    min: int,
    max: Optional[int] = None,
    *,
    predicate: Optional[Callable[[bytes], bool]] = None,
) -> bytearray:
    """
    Generates a random bytearray of length between min and max.
    If max is None, the length is exactly min.
    """

    def gen() -> bytearray:
        if max is None:
            len = min
        else:
            len = random.randint(min, max)
        return bytearray(random.getrandbits(8) for _ in range(len))

    ret = gen()
    if predicate is None:
        return ret

    while not predicate(ret):
        ret = gen()
    return ret


generators_map = {
    bool: lambda: random.choice([True, False]),
    Address: lambda: random_address(),
    Wei: lambda: Wei(random.randint(0, 10000000000000000000)),
}


def generate(t: Type, options: Optional[Dict[str, Any]] = None):
    if options is None:
        options = {}

    if "length" in options:
        min_len = options["length"]
        max_len = options["length"]
    else:
        min_len = 0
        max_len = 64

    try:
        return generators_map[t]()
    except KeyError:
        if get_origin(t) is Annotated:
            args = get_args(t)
            opt = {}

            for arg in args[1:]:
                if isinstance(arg, Length):
                    opt["length"] = arg.length
                elif isinstance(arg, ValueRange):
                    opt["min"] = arg.min
                    opt["max"] = arg.max

            return generate(args[0], opt)
        elif get_origin(t) is list:
            return [
                generate(get_args(t)[0])
                for _ in range(random.randint(min_len, max_len))
            ]
        elif t is int:
            min = options.get("min", -(2**255))
            max = options.get("max", 2**255 - 1)
            return random.randint(min, max)
        elif t is bytes:
            return bytes(random_bytes(min_len, max_len))
        elif t is bytearray:
            return random_bytes(min_len, max_len)
        elif t is str:
            return random_string(min_len, max_len)
        elif issubclass(t, enum.Enum):
            return random.choice(list(t))
        elif dataclasses.is_dataclass(t):
            return t(
                *[generate(h) for h in get_type_hints(t, include_extras=True).values()]
            )
        else:
            raise ValueError(f"No fuzz generator found for type {t}")



from __future__ import annotations

import logging
from typing import List, Set

import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)

from .reentrancy import address_is_safe, find_low_level_call_source_address

logger = logging.getLogger(__name__)


def check_delegatecall_in_function(
    function_definition: ir.FunctionDefinition,
    statement: ir.StatementAbc,
    address_source: ir.ExpressionAbc,
    checked_statements: Set[ir.StatementAbc],
) -> List[Detection]:
    from woke.analysis.ownable import statement_is_publicly_executable
    from woke.analysis.utils import pair_function_call_arguments

    if not statement_is_publicly_executable(statement):
        return []

    source_address_declaration = find_low_level_call_source_address(address_source)
    is_safe = None
    if source_address_declaration is None:
        logger.debug(f"{address_source.source}")
    elif isinstance(source_address_declaration, ir.enums.GlobalSymbolsEnum):
        if source_address_declaration == ir.enums.GlobalSymbolsEnum.THIS:
            is_safe = True
        elif source_address_declaration in {
            ir.enums.GlobalSymbolsEnum.MSG_SENDER,
            ir.enums.GlobalSymbolsEnum.TX_ORIGIN,
        }:
            is_safe = False
        else:
            is_safe = None
            logger.debug(f"{source_address_declaration}:")
    elif isinstance(source_address_declaration, ir.ContractDefinition):
        if source_address_declaration.kind == ir.enums.ContractKind.LIBRARY:
            is_safe = True
    elif isinstance(source_address_declaration, ir.Literal):
        is_safe = True
    else:
        is_safe = address_is_safe(source_address_declaration)

    if is_safe:
        return []

    checked_statements.add(statement)
    ret = []
    if function_definition.visibility in {
        ir.enums.Visibility.PUBLIC,
        ir.enums.Visibility.EXTERNAL,
    }:
        ret.append(
            Detection(
                statement,
                f"Exploitable from `{function_definition.canonical_name}`, address is safe: {is_safe}",
            )
        )

    for ref in function_definition.get_all_references(False):
        if isinstance(ref, ir.IdentifierPathPart):
            top_statement = ref.underlying_node
        elif isinstance(ref, ir.ExternalReference):
            continue  # TODO currently not supported
        else:
            top_statement = ref
        func_call = None
        while top_statement is not None:
            if (
                func_call is None
                and isinstance(top_statement, ir.FunctionCall)
                and top_statement.function_called == function_definition
            ):
                func_call = top_statement
            if isinstance(top_statement, ir.StatementAbc):
                break
            top_statement = top_statement.parent

        if top_statement is None or func_call is None:
            continue
        function_def = top_statement
        while function_def is not None and not isinstance(
            function_def, ir.FunctionDefinition
        ):
            function_def = function_def.parent
        if function_def is None:
            continue
        assert isinstance(function_def, ir.FunctionDefinition)
        if top_statement in checked_statements:
            continue

        if source_address_declaration in function_definition.parameters.parameters:
            for arg_decl, arg_expr in pair_function_call_arguments(
                function_definition, func_call
            ):
                if arg_decl == source_address_declaration:
                    assert isinstance(arg_expr.type, (types.Address, types.Contract))
                    ret.extend(
                        check_delegatecall_in_function(
                            function_def, top_statement, arg_expr, checked_statements
                        )
                    )
                    break
        else:
            ret.extend(
                check_delegatecall_in_function(
                    function_def, top_statement, address_source, checked_statements
                )
            )

    return ret


class UnsafeDelegatecallDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_member_access(self, node: ir.MemberAccess):
        t = node.type
        if (
            not isinstance(t, types.Function)
            or t.kind
            not in {
                ir.enums.FunctionTypeKind.DELEGATE_CALL,
                ir.enums.FunctionTypeKind.BARE_DELEGATE_CALL,
            }
            or t.attached_to is not None
        ):
            return

        address_source = node.expression
        statement = node
        while statement is not None:
            if isinstance(statement, ir.StatementAbc):
                break
            statement = statement.parent
        if statement is None:
            return
        func = statement
        while func is not None:
            if isinstance(func, ir.FunctionDefinition):
                break
            func = func.parent
        if func is None:
            return

        ret = check_delegatecall_in_function(func, statement, address_source, set())
        if len(ret) == 0:
            return

        self._detections.add(
            DetectorResult(
                Detection(
                    node,
                    f"Possibly unsafe delegatecall in `{func.canonical_name}`",
                    tuple(ret),
                ),
                confidence=DetectionConfidence.MEDIUM,
                impact=DetectionImpact.HIGH,
            )
        )

    @detector.command("unsafe-delegatecall")
    def cli(self):
        """
        Delegatecall to an untrusted contract.
        """
        pass



from __future__ import annotations

import logging
import re
from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from woke.ir.ast import SolcEnumDefinition

from ..abc import IrAbc, SolidityAbc
from ..meta.structured_documentation import StructuredDocumentation
from ..utils import IrInitTuple
from .abc import DeclarationAbc
from .enum_value import EnumValue

if TYPE_CHECKING:
    from ..meta.source_unit import SourceUnit
    from .contract_definition import ContractDefinition


logger = logging.getLogger(__name__)


class EnumDefinition(DeclarationAbc):
    """
    Definition of an enum.

    !!! example
        ```solidity
        enum ActionChoices { GoLeft, GoRight, GoStraight, SitStill }
        ```
    """

    _ast_node: SolcEnumDefinition
    _parent: Union[ContractDefinition, SourceUnit]

    _canonical_name: str
    _values: List[EnumValue]
    _documentation: Optional[StructuredDocumentation]

    def __init__(
        self, init: IrInitTuple, enum: SolcEnumDefinition, parent: SolidityAbc
    ):
        super().__init__(init, enum, parent)
        self._canonical_name = enum.canonical_name

        self._values = []
        for value in enum.members:
            self._values.append(EnumValue(init, value, self))
        self._documentation = (
            StructuredDocumentation(init, enum.documentation, self)
            if enum.documentation is not None
            else None
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for value in self._values:
            yield from value

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        ENUM_RE = re.compile(
            r"^\s*enum\s+(?P<name>{identifier})".format(identifier=IDENTIFIER).encode(
                "utf-8"
            )
        )

        byte_start = self._ast_node.src.byte_offset
        match = ENUM_RE.match(self._source)
        assert match
        return byte_start + match.start("name"), byte_start + match.end("name")

    @property
    def parent(self) -> Union[SourceUnit, ContractDefinition]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return self._canonical_name

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        return (
            f"enum {self.name}"
            + " {\n"
            + ",\n".join(f"    {value.name}" for value in self._values)
            + "\n}"
        )

    @property
    def values(self) -> Tuple[EnumValue, ...]:
        """
        Returns:
            Enum values defined in this enum.
        """
        return tuple(self._values)

    @property
    def documentation(self) -> Optional[StructuredDocumentation]:
        """
        Added in Solidity 0.8.20.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation



import lazy_import

lazy_import.lazy_module("networkx")
lazy_import.lazy_module("woke.ir")
lazy_import.lazy_module("woke.ir.enums")
lazy_import.lazy_module("woke.ir.types")
lazy_import.lazy_module("woke.analysis")
lazy_import.lazy_module("woke.analysis.cfg")
lazy_import.lazy_module("woke.analysis.ownable")
lazy_import.lazy_module("woke.analysis.utils")
